# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import time
import sys

import jax
from jax import lax, config
import jax.numpy as jnp
import flax

from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, spaces

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')


# Dirichlet problem of the Poisson equation in 2d: [0,1]x[0,1]
# Demonstration of compile time reduction by manually pre-compiling solution structure


# Positive smooth distance functions that select the specific boundary segments
psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, [0.,0.], [1.,0.])
psdf_right = lambda x: geometry.psdf_trimmed_line(x, [1.,0.], [1.,1.])
psdf_top = lambda x: geometry.psdf_trimmed_line(x, [1.,1.], [0.,1.])
psdf_left = lambda x: geometry.psdf_trimmed_line(x, [0.,1.], [0.,0.])
psdf_rect = lambda x: geometry.psdf_parallelogram(x, [0.,0.], [0.,1.], [1.,0.])
psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left]

# Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
psdf_domain = psdf_rect

# Boundary condition functions
def homogeneous(x): # Bottom, right and top boundaries
  return 0

# Assignment of boundary conditions
dirichlet_conditions = [homogeneous, homogeneous, homogeneous, homogeneous]


### Transfinite interpolation of boundary segments
b_cond = lambda x: geometry.transfinite_interpolation(x, psdfs, dirichlet_conditions)
psdf = psdf_domain

### Initialization and generation of nodes and integration points
# Regular node placement
nod_spacing = 0.02
(x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_domain, [0.,0.], [1.,1.], nod_spacing)

# Integration points
# Collocation, Note: this can in some cases cause bad conditioning, unphysical low-energy modes and divergence. 
# If so, use more integration points, e.g. Gauss integration on a background grid (seeder.gauss_points_in_psdf)
x_int = x_nodes
n_int = n_nodes

# Support radius for moving least square ansatz
support_radius = 2.51 * nod_spacing

# Integration point weights
w_int = (region_size / n_int) * jnp.ones(n_int)

# Definition of pde
pde = models.poisson(source_fun = lambda x: 1)


### Neighborhood search
(num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


### Settings
n_fields = 1
dofs0 = jnp.zeros((n_nodes, n_fields))
static_settings = flax.core.FrozenDict({

  # Precompilation of the solution structure at the integration points can considerable improve compile time
  # 'shape function mode': 'direct',
  'shape function mode': 'compiled', # precompilation is done in line 107
  
  'solution space': ('mls',),
  'number of fields': (n_fields,),
  'assembling mode': ('sparse',),
  'maximal number of neighbors': (max_neighbors,),
  'order of basis functions': (2,),
  'weight function type': ('gaussian',),
  'variational scheme': ('strong form galerkin',),
  'model': (pde,),
  'solution structure': ('dirichlet',),
  'boundary conditions': (b_cond,),
  'psdf': (psdf,),
  'solver type': 'linear',

  'solver backend': 'pardiso',
  'solver': 'lu',
  # 'solver backend': 'jax',
  # 'solver': 'bicgstab',
  # 'type of preconditioner': 'jacobi',
  # 'hvp type': 'fwdrev',

  'verbose': 2,
  'connectivity': (utility.jnp_to_tuple(neighbor_list),),
})
settings = {
  'node coordinates': x_nodes,
  'beta': (3,),
  'integration coordinates': (x_int,),
  'integration weights': (w_int,),
  'support radius': (support_radius,),
}

### Precompilation of solution structure
start = time.time()
settings = solution_structures.precompile(dofs0, settings, static_settings)


### Call solver
dofs, _ = solver.solver(dofs0, settings, static_settings)
print("Time needed for compiling and solving: ", time.time() - start)

### Preperation of postprocessing data
local_dofs = dofs[neighbor_list]
post_fun_vj = jax.jit(jax.vmap(solution_structures.solution_structure, (0,0,0,None,None,None)), static_argnames=['static_settings', 'set'])
data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings, 0)

# Plotting with pyvista
plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True)



# solver.py
# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
This module is the central module of the analysis phase.
Based on the given entries in settings and static_settings, the functions solver and adaptive_load_stepping can be used to find the roots of the global residual vector. 
Depending on the settings, linear equation solvers, the Newton-Raphson method, or nonlinear minimizers are utilized. 
The residual vectors and (in the case of external solvers) the tangent matrix are automatically assembled according to the chosen settings. 
The solver module uses the assembler, which in turn calls the variational_scheme, the solution_structure, and the spaces modules. 
Additionally, automatic implicit differentiation in forward or reverse mode via the implicit_diff module is provided for the adaptive_load_stepping function.
For solving the linear equation systems, wrappers for different backends on CPU and GPU are available, including Pardiso and PETSc.
"""

import time
import sys

import jax
import jaxopt
import jax.numpy as jnp
from jax.experimental import sparse
from jax import lax
import numpy as np
import scipy as scp

from autopdex import assembler, implicit_diff
from autopdex.utility import jit_with_docstring



### Solvers as specified by the static_settings and settings

@jit_with_docstring(static_argnames=['static_settings', '**kwargs'])
def solver(dofs, settings, static_settings, **kwargs):
  """
  General solver function to solve a given problem based on provided settings.

  This function chooses and runs the appropriate solver type (e.g., minimization, linear, Newton) based on the provided
  `static_settings` and returns the solution and any additional information.

  Parameters:
      dofs (jnp.ndarray): Degrees of freedom or initial guess for the solution.
      settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
      static_settings (dict): Dictionary containing static settings such as solver type, verbose level, and variational schemes.
      **kwargs (dict): Additional keyword arguments passed to the specific solver functions.

  Returns:
      jnp.ndarray: The solution obtained from the selected solver.
      Any: Additional information from the solver, such as number of iterations or convergence status.

  Solver Types:
    - 'minimize' : Uses nonlinear minimization solvers (e.g., LBFGS, BFGS, etc.).
    - 'linear' : Solves linear systems using specified backend (e.g., JAX, PETSc, PARDISO, PyAMG, Scipy).
    - 'diagonal linear' : Solves linear systems assuming a diagonal tangent matrix.
    - 'newton' : Uses the Newton method for solving nonlinear systems.
    - 'damped newton' : Uses a damped Newton method for solving nonlinear systems.

  For the different solvers, the function conducts the following functions in which more documentation is provided:
    - 'minimize' : solver.solve_nonlinear_minimization
    - 'linear' : solver.solve_linear
    - 'diagonal linear' : solver.solve_diagonal_linear
    - 'newton' : solver.solve_newton
    - 'damped newton' : solver.solve_damped_newton

  Notes:
    - If all domains are using 'least square pde loss' variational scheme and verbosity level is >=1, it prints the L2 error before and after optimization.
  """
  
  # Give global error estimator with unfitted dofs if all domains are 'least square pde loss'
  verbose = static_settings['verbose']
  try:
    if all([scheme == 'least square pde loss' for scheme in static_settings['variational scheme']]) and verbose>=1:
      jax.debug.print("L2 error unoptimized: {x}", x=assembler.integrate_functional(dofs, settings, static_settings))
  except KeyError:
    pass

  # Choose type of solver and call it
  solver_type = static_settings['solver type']
  match solver_type:
    case 'minimize':
      sol = solve_nonlinear_minimization(dofs, settings, static_settings, **kwargs)
      infos = None
    case 'linear':
      sol = solve_linear(dofs, settings, static_settings, **kwargs)
      infos = None
    case 'diagonal linear':
      sol = solve_diagonal_linear(dofs, settings, static_settings, **kwargs)
      infos = None
    # case 'diagonal newton':
    #   # ToDo
    #   sol, infos = solve_diagonal_newton(dofs, settings, static_settings, **kwargs)
    case 'newton':
      sol, infos = solve_newton(dofs, settings, static_settings, **kwargs)
    case 'damped newton':
      sol, infos = solve_damped_newton(dofs, settings, static_settings, **kwargs)
    case _:
      assert False, 'Solver type not implemented!'

  # Give global error estimator with fitted dofs if all domains are 'least square pde loss'
  try:
    if all([scheme == 'least square pde loss' for scheme in static_settings['variational scheme']]) and verbose>=1:
      jax.debug.print("L2 error optimized: {x}", x=assembler.integrate_functional(sol, settings, static_settings))
  except KeyError:
    pass
  return sol, infos

@jit_with_docstring(static_argnames=['static_settings', 'multiplier_settings', 'path_dependent', 'implicit_diff_mode', 'max_multiplier', 'min_increment', 'max_increment', 'init_increment', 'target_num_newton_iter', 'max_load_steps', 'newton_tol', '**kwargs'])
def adaptive_load_stepping(dofs, settings, static_settings, multiplier_settings = lambda settings, multiplier: (settings.update({'load multiplier': multiplier}), settings)[1], path_dependent = True, implicit_diff_mode = None, max_multiplier = 1.0, min_increment = 0.01, max_increment = 1.0, init_increment = 0.2, max_load_steps = 1000, target_num_newton_iter = 7, newton_tol = 1e-10, **kwargs):
  """
  Performs adaptive load stepping to solve a nonlinear system of equations.

  This function iteratively adjusts the load increment to ensure convergence 
  using a Newton-Raphson solver. The increment size is adaptively controlled 
  based on the convergence behavior of the solver. Works currently only with
  solver types 'newton' and 'damped newton'. 

  Args:
    dofs (jnp.ndarray): Initial degrees of freedom.
    settings (dict): Dictionary of problem settings.
    static_settings (dict): Dictionary of static settings that do not change during load steps.
    multiplier_settings (callable): Function to update settings based on the current load multiplier.
    path_dependent (bool): Specifies wether problem is path-dependent (experimental) or not (has an influence on the implicit differentiation).
    implicit_diff_mode (string): Can be either \'reverse\', \'forward\' or None. In case of \'reverse\', only reverse mode differentiation is supported (jacrev), in case of \'forward\', only forward mode differentiation is supported (jacfwd).
    max_multiplier (float): Maximum value for the load multiplier.
    min_increment (float): Minimum allowable increment size.
    max_increment (float): Maximum allowable increment size.
    init_increment (float): Initial increment size.
    max_load_steps (int): Maximal number of load steps. Only used in case implicit_diff_mod is not None
    target_num_newton_iter (int): Target number of Newton iterations for each load step.
    newton_tol (float, optional): Tolerance for Newton solver convergence. Default is 1e-10.
    **kwargs: Additional keyword arguments for the solver.

  Returns:
      - jnp.ndarray: Solution degrees of freedom after load stepping.
  """
  verbose = static_settings['verbose']

  if implicit_diff_mode is not None:
    # Set-up the decorator for implicit differentiation
    residual_fun = lambda dofs, settings: assembler.assemble_residual(dofs, settings, static_settings)
    tangent_fun = lambda dofs, settings: assembler.assemble_tangent(dofs, settings, static_settings)
    free_dofs_flat = None
    try:
      dirichlet_dofs_flat = np.asarray([element for tupl in static_settings['dirichlet dofs'] for element in tupl])
      free_dofs_flat = np.invert(dirichlet_dofs_flat)
    except KeyError:
      pass

    solver_backend = static_settings['solver backend']
    solver_subtype = static_settings['solver']
    try: 
      sensitivity_solver_backend = static_settings['sensitivity solver backend']
      sensitivity_solver_subtype = static_settings['sensitivity solver']
    except KeyError:
      sensitivity_solver_backend = solver_backend
      sensitivity_solver_subtype = solver_subtype

    match sensitivity_solver_backend:
      case 'petsc':
        n_fields = static_settings['number of fields']
        pc_type = static_settings['type of preconditioner']
        lin_solve_fun = lambda mat, rhs: linear_solve_petsc(mat, rhs, n_fields, sensitivity_solver_subtype, pc_type, verbose, free_dofs_flat, **kwargs)
      case 'pardiso':
        lin_solve_fun = lambda mat, rhs: linear_solve_pardiso(mat, rhs, sensitivity_solver_subtype, verbose, free_dofs_flat, **kwargs)
      case 'pyamg':
        pc_type = static_settings['type of preconditioner']
        lin_solve_fun = lambda mat, rhs: linear_solve_pyamg(mat, rhs, sensitivity_solver_subtype, pc_type, verbose, free_dofs_flat, **kwargs)
      case 'scipy':
        lin_solve_fun = lambda mat, rhs: linear_solve_scipy(mat, rhs, sensitivity_solver_subtype, verbose, free_dofs_flat, **kwargs)
      case _:
          raise ValueError("Specified sensitivity solver backend not available. Choose \'pardiso\', \'petsc\', \'pyamg\' or \'scipy\'.")
    lin_solve_callback_fun = lambda mat, rhs: jax.pure_callback(lin_solve_fun, jnp.zeros(rhs.shape, rhs.dtype), mat, rhs)

  # Set up functions for adaptive load stepping loop
  def continue_check(carry):
    _, multiplier, increment, _, _, _ = carry
    _continue_1 = jnp.logical_and(multiplier < max_multiplier, increment > min_increment)
    jax.lax.cond(jnp.logical_and(multiplier < max_multiplier - min_increment, increment < min_increment), 
                 lambda: jax.debug.print("Adaptive load stepping could not converge; increment size below min_increment!"),
                 lambda: None)
    return _continue_1
  
  def step(carry):
    dofs0, multiplier, increment, load_step, settings, _ = carry

    # Update multiplier
    multiplier += increment

    # Update boundary conditions
    if verbose > -1:
      if verbose > 0:
        jax.debug.print('')
      jax.debug.print('Multiplier: {x}', x=multiplier)
    settings = multiplier_settings(settings, multiplier)

    # Call newton solver
    if path_dependent and implicit_diff_mode is not None: # Add implicit differentiation for each load step
      @implicit_diff.custom_root(residual_fun, tangent_fun, lin_solve_callback_fun, free_dofs_flat, True, implicit_diff_mode)
      def diffable_solve(dofs0, settings):
        return solver(dofs0, settings, static_settings, newton_tol=newton_tol, **kwargs)
      dofs, infos = diffable_solve(dofs0, settings)
      needed_steps, res_norm_free_dofs, divergence = infos

    else: # Add implicit diff wrapper on the adaptive load stepping level
      dofs, infos = solver(dofs0, settings, static_settings, newton_tol=newton_tol, **kwargs)
      needed_steps, res_norm_free_dofs, divergence = infos

    # Adaptive incrementation
    multiplier = jnp.where(divergence, multiplier - increment, multiplier)
    increment = jnp.where(divergence, 0.5 * increment,
                          (1 + 0.5 * (target_num_newton_iter - needed_steps) / target_num_newton_iter) * increment)
    increment = jnp.where(increment > max_increment, max_increment, increment)

    dofs = jnp.where(divergence, dofs0, dofs)

    # Limit to max_multiplier
    increment = jnp.where(multiplier + increment > max_multiplier, max_multiplier - multiplier, increment)
    return dofs, multiplier, increment, 1.*load_step, settings, res_norm_free_dofs

  # Use implicit diff wrappers to make it differentiable
  if implicit_diff_mode is not None:
    if not path_dependent: # Conservative problem

      # Set Dirichlet conditions for derivative w.r.t. them
      settings = multiplier_settings(settings, max_multiplier)

      @implicit_diff.custom_root(residual_fun, tangent_fun, lin_solve_callback_fun, free_dofs_flat, True, implicit_diff_mode)
      def diffable_adaptive_load_stepping(dofs, settings):
        return jax.lax.while_loop(continue_check, step, (dofs, 0., init_increment, 0, settings, 0.))
      return diffable_adaptive_load_stepping(dofs, settings)

    else: # Pathdependent problem; uses fori_loop with static limits for supporting reverse mode differentiation
      def body_fn(i, carry):
        def continue_check(carry):
          _, multiplier, increment, _, _, divergence, stop = carry
          _continue_1 = jnp.logical_and(multiplier < max_multiplier, increment > min_increment)
          jax.lax.cond(jnp.logical_and(jnp.logical_and(multiplier < max_multiplier, increment < min_increment), jnp.logical_not(stop)),
                      lambda: jax.debug.print("Adaptive load stepping could not converge; increment size below min_increment!"),
                      lambda: None)
          return _continue_1

        def step_extended(carry):
          dofs, multiplier, increment, load_step, settings, res_norm_free_dofs, stop = carry
          args = dofs, multiplier, increment, load_step, settings, res_norm_free_dofs
          return step(args) + (False,)

        def finish(carry):
          dofs, multiplier, increment, load_step, settings, res_norm_free_dofs, stop = carry
          return dofs, multiplier, increment, load_step, settings, res_norm_free_dofs, True

        carry = jax.lax.cond(continue_check(carry), lambda x: step_extended(x), lambda x: finish(x), carry)

        # ToDo: Verify accuracy of derivatives with finite differences.
        return carry
      init_state = (dofs, 0., init_increment, 0, settings, 0., False)
      return jax.lax.fori_loop(0, max_load_steps, body_fn, init_state)

  else: # No definition of implicit derivatives
    return jax.lax.while_loop(continue_check, step, (dofs, 0., init_increment, 0, settings, 0.))


### Minimizers
@jit_with_docstring(static_argnames=['static_settings', '**kwargs'])
def solve_nonlinear_minimization(dofs, settings, static_settings, **kwargs):
  """
  Solves a nonlinear minimization problem using specified optimization methods.

  This function wraps nonlinear minimization solvers provided by `jaxopt` to minimize a functional
  and solve the given problem.

  Parameters:
    dofs (jnp.ndarray): Degrees of freedom or initial guess for the solution.
    settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
    static_settings (dict): Dictionary containing static settings such as solver type, verbose level, and variational schemes.
    **kwargs (dict): Additional keyword arguments passed to the specific solver functions.

  Returns:
    jnp.ndarray: The optimized solution obtained from the selected solver.

  Solver Types
    - 'gradient descent' : Uses gradient descent for optimization.
    - 'lbfgs' : Uses Limited-memory Broyden-Fletcher-Goldfarb-Shanno (LBFGS) algorithm for optimization.
    - 'bfgs' : Uses Broyden-Fletcher-Goldfarb-Shanno (BFGS) algorithm for optimization.
    - 'nonlinear cg' : Uses nonlinear conjugate gradient method for optimization.
    - 'gauss newton' : Uses Gauss-Newton method for optimization.
    - 'levenberg marquart' : Uses Levenberg-Marquardt algorithm for optimization.
    - Default : If solver name is not set or not available, uses 'lbfgs' as the default solver.

  Notes:
    - This function should just be called, if the variational scheme involves the definition of a functional that is to be minimized,
      e.g. 'least square pde loss'. The modes 'gauss newton' and 'levenberg marquart' are an exeption, since the utilize the residual.
    - The function conducts the assembler.integrate_functional and assembler.assemble_residual functions in order
      to set up suitable optimization functions or residual functions, depending on what the solver needs.
    - The current implementation does not support nodal imposition of DOFs. This will raise an assertion error if
      nodal imposition is detected in the `static_settings`.
  """
  nodal_imposition = 'nodal imposition' in static_settings['solution structure']
  assert not nodal_imposition, "solver type \'minimize\' does currently not support nodal imposition of DOFs."
  # ToDo: impose boundary conditions and freeze dirichlet dofs

  def functional(params):
    return assembler.integrate_functional(params, settings, static_settings)
  def residual_function(params):
    return assembler.assemble_residual(params, settings, static_settings)
  def residual_function_flat(params):
    return assembler.assemble_residual(jnp.reshape(params, dofs.shape), settings, static_settings).flatten()

  # Select minimizer
  solver_name = static_settings['solver']
  match solver_name:
    case 'gradient descent':
      solver = jaxopt.GradientDescent(functional, **kwargs)
      sol = solver.run(dofs).params
    case 'lbfgs':
      solver = jaxopt.LBFGS(functional, **kwargs)
      sol = solver.run(dofs).params
    case 'bfgs':
      solver = jaxopt.BFGS(functional, **kwargs)
      sol = solver.run(dofs).params
    case 'nonlinear cg':
      solver = jaxopt.NonlinearCG(functional, **kwargs)
      sol = solver.run(dofs).params
    case 'gauss newton':
      solver = jaxopt.GaussNewton(residual_function, **kwargs)
      sol = solver.run(dofs).params
    case 'levenberg marquart':
      solver = jaxopt.LevenbergMarquardt(residual_function_flat, **kwargs)
      sol = jnp.reshape(solver.run(dofs.flatten()).params, dofs.shape)
    case _:
      solver = jaxopt.LBFGS(functional, **kwargs)
      sol = solver.run(dofs).params
      print('Solver name not set or not available in combination with this solver type. Using static_settings[\'solver name\'] = \'lbfgs\' as default.')
  return sol


### Root finders
@jit_with_docstring(static_argnames=['static_settings', '**kwargs'])
def solve_linear(dofs, settings, static_settings, **kwargs):
  """
  Solves a linear system using the specified backend and solver settings.

  This function determines the appropriate linear solver based on the provided settings and forwards the
  call to the selected solver function. It supports both JAX matrix-free solvers and external solvers
  like PETSc, PARDISO, PyAMG, and Scipy using jax.pure_callback.

  Parameters:
    dofs (jnp.ndarray): Degrees of freedom or initial guess for the solution.
    settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
    static_settings (dict): Dictionary containing static settings such as solver backend, type of solver, and preconditioner.
    **kwargs (dict): Additional keyword arguments passed to the specific solver functions.

  Returns:
    jnp.ndarray: The solution obtained from the selected linear solver.

  Solver Backends:
    - 'jax' : Uses JAX's matrix-free solvers.
    - 'petsc' : Uses PETSc for solving linear systems.
    - 'pardiso' : Uses PARDISO for solving linear systems.
    - 'pyamg' : Uses PyAMG for solving linear systems.
    - 'scipy' : Uses Scipy's sparse solvers for solving linear systems.

  Notes:
    - If `nodal imposition` is detected in the `static_settings`, the function imposes Dirichlet boundary conditions
      and adjusts the degrees of freedom accordingly.
    - The function assembles the residual and tangent matrix before solving the system in case an external solver is used.
    - If the tangent matrix is dense and an external solver is used, it is converted to a sparse format.
    - The function uses JAX's `pure_callback` to call external solvers and handle the solution.
  """
  solver_backend = static_settings['solver backend']

  ### Jax matrix-free solver
  if solver_backend == 'jax':
    return linear_solve_jax(dofs, settings, static_settings, **kwargs)

  ### External linear solver
  nodal_imposition = 'nodal imposition' in static_settings['solution structure']
  if nodal_imposition:
    # Impose Dirichlet boundaries. Dirichlet dofs has to be concrete, therefore it is passed in static_settings as tuple of tuples
    dirichlet_dofs_flat = np.asarray([element for tupl in static_settings['dirichlet dofs'] for element in tupl])
    free_dofs_flat = np.invert(dirichlet_dofs_flat)

    # Impose nodal dofs
    dirichlet_conditions = settings['dirichlet conditions'].flatten()
    flat_dofs = dofs.flatten()
    idx = jnp.arange(flat_dofs.shape[0])[dirichlet_dofs_flat]
    flat_dofs = flat_dofs.at[idx].set(dirichlet_conditions[idx])
    dofs = flat_dofs.reshape(dofs.shape)

  # Assembling
  verbose = static_settings['verbose']
  solver = static_settings['solver']
  rhs = - assembler.assemble_residual(dofs, settings, static_settings).flatten()
  mat = assembler.assemble_tangent(dofs, settings, static_settings)

  # If dense matrix and external solver, convert to sparse
  if solver_backend in ('petsc', 'pardiso', 'pyamg', 'scipy') and type(mat) == jnp.ndarray:
    mat = sparse.bcoo_fromdense()

  if nodal_imposition:
    # Delete rows
    rhs = rhs[free_dofs_flat]

  match solver_backend:
    case 'petsc':
      n_fields = static_settings['number of fields']
      pc_type = static_settings['type of preconditioner']
      solve_fun = lambda a, b, c : linear_solve_petsc(a, b, n_fields, solver, pc_type, verbose, free_dofs=c, **kwargs)
    case 'pardiso':
      solve_fun = lambda a, b, c: linear_solve_pardiso(a, b, solver, verbose, free_dofs=c)
    case 'pyamg':
      pc_type = static_settings['type of preconditioner']
      solve_fun = lambda a, b, c: linear_solve_pyamg(a, b, solver, pc_type, verbose, free_dofs=c, **kwargs)
    case 'scipy':
      solve_fun = lambda a, b, c: linear_solve_scipy(a, b, solver, verbose, free_dofs=c)

  # Prepare callback
  result_shape_dtype = jax.ShapeDtypeStruct(shape=rhs.shape, dtype=rhs.dtype)

  if nodal_imposition:
    sol = jax.pure_callback(solve_fun, result_shape_dtype, mat, rhs, free_dofs_flat)

    # Compose solution dofs
    idx = jnp.arange(flat_dofs.shape[0])[free_dofs_flat]
    flat_dofs = flat_dofs.at[idx].set(sol)
    dofs = flat_dofs.reshape(dofs.shape)
    return dofs
  else:
    return jax.pure_callback(solve_fun, result_shape_dtype, mat, rhs, None).reshape(dofs.shape)

@jit_with_docstring(static_argnames=['static_settings', '**kwargs'])
def solve_diagonal_linear(dofs, settings, static_settings, **kwargs):
  """
  Solves a linear system assuming the tangent matrix is diagonal.

  This function solves the linear system by leveraging the assumption that the tangent matrix is diagonal,
  which simplifies the solution process. It supports nodal imposition of Dirichlet boundary conditions and
  handles the assembly of the residual and diagonal tangent matrix. 
  
  If the tangent matrix is not diagonal, it will produce a wrong diagonal of the tangent!

  Parameters:
    dofs (jnp.ndarray): Degrees of freedom or initial guess for the solution.
    settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
    static_settings (dict): Dictionary containing static settings such as solution structure and solver backend.
    **kwargs (dict): Additional keyword arguments passed to the solver function.

  Returns:
    sol (jnp.ndarray): The solution obtained by solving the linear system assuming a diagonal tangent matrix.

  Notes:
    - If `nodal imposition` is detected in the `static_settings`, the function imposes Dirichlet boundary conditions
      and adjusts the degrees of freedom accordingly.
    - The function assembles the residual and diagonal tangent matrix before solving the system.
    - The solution process involves element-wise division of the residual by the diagonal elements of the tangent matrix.
  """
  nodal_imposition = 'nodal imposition' in static_settings['solution structure']
  if nodal_imposition:
    # Impose Dirichlet boundaries. Dirichlet dofs has to be concrete, therefore it is passed in static_settings as tuple of tuples
    dirichlet_dofs_flat = np.asarray([element for tupl in static_settings['dirichlet dofs'] for element in tupl])
    free_dofs_flat = np.invert(dirichlet_dofs_flat)

    # Impose nodal dofs
    dirichlet_conditions = settings['dirichlet conditions'].flatten()
    flat_dofs = dofs.flatten()
    idx = jnp.arange(flat_dofs.shape[0])[dirichlet_dofs_flat]
    flat_dofs = flat_dofs.at[idx].set(dirichlet_conditions[idx])
    dofs = flat_dofs.reshape(dofs.shape)

  # Assembling
  rhs = - assembler.assemble_residual(dofs, settings, static_settings).flatten()
  diag_mat = assembler.assemble_tangent(dofs, settings, static_settings).data

  if nodal_imposition:
    # Delete rows
    rhs = rhs[free_dofs_flat]
    diag_mat = diag_mat[free_dofs_flat]

  # Solve while assuming a diagonal tangent
  sol = jnp.multiply((1/diag_mat), rhs)

  if nodal_imposition:
    # Compose solution dofs
    idx = jnp.arange(flat_dofs.shape[0])[free_dofs_flat]
    flat_dofs = flat_dofs.at[idx].set(sol)
    dofs = flat_dofs.reshape(dofs.shape)
    return dofs
  else:
    return sol.reshape(dofs.shape)

@jit_with_docstring(static_argnames=['static_settings', '**kwargs'])
def solve_newton(dofs, settings, static_settings, newton_tol=1e-8, maxiter=30, **kwargs):
  """
  Solves a nonlinear system using the Newton-Raphson method.

  This function is a wrapper for the damped Newton method with a damping coefficient of 1.0,
  effectively performing standard Newton-Raphson iterations.

  Parameters:
    dofs (jnp.ndarray): Degrees of freedom or initial guess for the solution.
    settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
    static_settings (dict): Dictionary containing static settings such as solution structure and solver backend.
    newton_tol (float, optional): Tolerance for the Newton method convergence criterion. Default is 1e-8.
    maxiter (int, optional): Maximum number of iterations for the Newton method. Default is 30.
    **kwargs (dict): Additional keyword arguments passed to the solver function.

  Returns:
    tuple: A tuple containing the following elements:
      - sol (jnp.ndarray): The solution obtained by solving the nonlinear system using the Newton method.
      - infos (tuple): Additional information about the solution process, including:
          - num_iterations (int): The number of iterations performed.
          - residual_norm (float): The norm of the residual at the solution.
          - diverged (bool): Flag indicating whether the method diverged.
  """
  return solve_damped_newton(dofs, settings, static_settings, newton_tol, 1.0, maxiter)

@jit_with_docstring(static_argnames=['static_settings', '**kwargs'])
def solve_damped_newton(dofs, settings, static_settings, newton_tol=1e-8, damping_coefficient=0.8, maxiter=30, **kwargs):
  """
  Solves a nonlinear system using the damped Newton method.

  This function performs damped Newton iterations to solve a nonlinear system, 
  with support for nodal imposition of Dirichlet boundary conditions specified as a boolean tuple-tree in static_settings['dirichlet dofs'].

  In this function the information needed for solver.damped_newton is prepared and the function is then called.

  Parameters:
    dofs (jnp.ndarray): Degrees of freedom or initial guess for the solution.
    settings (dict): Dictionary containing various settings and parameters required for assembling the problem.
    static_settings (dict): Dictionary containing static settings such as solution structure and solver backend.
    newton_tol (float, optional): Tolerance for the Newton method convergence criterion. Default is 1e-8.
    damping_coefficient (float, optional): Damping coefficient for the Newton updates. Default is 0.8.
    maxiter (int, optional): Maximum number of iterations for the Newton method. Default is 30.
    **kwargs (dict): Additional keyword arguments passed to the solver function.

  Returns:
    tuple: A tuple containing the following elements:
      - sol (jnp.ndarray): The solution obtained by solving the nonlinear system using the damped Newton method.
      - infos (tuple): Additional information about the solution process, including:
          - num_iterations (int): The number of iterations performed.
          - residual_norm (float): The norm of the residual at the solution.
          - diverged (bool): Flag indicating whether the method diverged.
  """
  residual_fun = lambda x_i: assembler.assemble_residual(x_i, settings, static_settings)
  lin_solve_fun = lambda x_i: solve_linear(x_i, settings, static_settings, **kwargs)

  nodal_imposition = 'nodal imposition' in static_settings['solution structure']
  if nodal_imposition:
    # Impose Dirichlet boundaries. Dirichlet dofs has to be concrete, therefore it is passed in static_settings as tuple of tuples
    dirichlet_dofs_flat = np.asarray([element for tupl in static_settings['dirichlet dofs'] for element in tupl])
    free_dofs_flat = np.invert(dirichlet_dofs_flat)
  else:
    free_dofs_flat = np.ones(dofs.flatten().shape, dtype=np.bool_)

  verbose = static_settings['verbose']
  return damped_newton(dofs, residual_fun, lin_solve_fun, free_dofs_flat, newton_tol, maxiter, damping_coefficient, verbose=verbose)

def damped_newton(dofs_0, residual_fun, lin_solve_fun, free_dofs_flat, newton_tol, maxiter, damping_coefficient, verbose=1):
  """
  Performs damped Newton iterations to solve a nonlinear system.

  This function implements the damped Newton method for solving a nonlinear system. It updates
  the solution iteratively based on the residual and chosen linear solver for the Newton step,
  with a damping coefficient to control the step size.

  Parameters:
    dofs_0 (jnp.ndarray): Initial guess for the degrees of freedom.
    residual_fun (callable): Function to compute the residual of the system.
    lin_solve_fun (callable): Function to solve the linearized system for the Newton step.
    free_dofs_flat (jnp.ndarray): Boolean array indicating the free degrees of freedom.
    newton_tol (float): Tolerance for the Newton method convergence criterion.
    maxiter (int): Maximum number of iterations for the Newton method.
    damping_coefficient (float): Damping coefficient for the Newton updates.

  Returns:
    tuple: A tuple containing the following elements:
      - sol (jnp.ndarray): The solution obtained by solving the nonlinear system using the damped Newton method.
      - infos (tuple): Additional information about the solution process, including:
          - num_iterations (int): The number of iterations performed.
          - residual_norm (float): The norm of the residual at the solution.
          - diverged (bool): Flag indicating whether the method diverged.
  """
  def step(carry):
    dofs_i, itt, _, res_norm_old, _ = carry
    
    # Update formula of newton scheme
    delta_x_i = lin_solve_fun(dofs_i)
    delta_x_i_flat = delta_x_i.flatten()

    # Update free dofs
    dofs_i_flat = dofs_i.flatten()
    idx_free = jnp.arange(delta_x_i_flat.shape[0])[free_dofs_flat]
    dofs_i_flat = dofs_i_flat.at[idx_free].add(damping_coefficient * delta_x_i_flat.at[idx_free].get())

    # Set boundary conditions
    idx_constrained = jnp.arange(delta_x_i_flat.shape[0])[np.invert(free_dofs_flat)]
    dofs_i_flat = dofs_i_flat.at[idx_constrained].set(delta_x_i_flat.at[idx_constrained].get())
    dofs_i = dofs_i_flat.reshape(dofs_i.shape)

    # Compute residual in next step as convergence test
    residual = residual_fun(dofs_i)
    residual_flat = residual.flatten()[free_dofs_flat]
    res_norm = jnp.linalg.norm(residual_flat)
    not_stop = jnp.where(res_norm > newton_tol, True, False)

    def report():
      if verbose > 0:
        jax.debug.print('Residual after Newton itteration {x}: {y}', x=itt+1, y=res_norm)     
        if verbose > 1:
          jax.debug.print('')

      # Check convergence
      divergence = jnp.where(jnp.logical_and(res_norm/res_norm_old > 10, itt > 1), True, False)
      return jnp.invert(divergence), divergence

    def stop_newton():
      jax.debug.print('Warning: Newton scheme could not converge!')
      return False, True

    next_step, divergence = jax.lax.cond(itt < maxiter, report, stop_newton)
    return (dofs_i, itt+1, jnp.logical_and(not_stop, next_step), res_norm, divergence)

  def convergence_check(carry):
    _, _, not_stop, _, _ = carry
    return not_stop
  sol, load_steps, _, res_norm, divergence  = lax.while_loop(convergence_check, step, (dofs_0, 0, True, 0., False))

  return (sol, (load_steps, res_norm, divergence))


### Linear solvers for different backends
@jit_with_docstring(static_argnames=['static_settings', '**kwargs'])
def linear_solve_jax(dofs, settings, static_settings, **kwargs):
  """
  Solves a linear system of equations using JAX's matrix free solvers.

  This function performs a linear solve using different itterative solvers, 
  optionally imposing Dirichlet boundary conditions and preconditioning and 
  using different matrix or Hessian vector product (HVP) methods.

  Args:
    dofs (jnp.ndarray): Initial degrees of freedom.
    settings (dict): Dictionary of problem settings.
    static_settings (dict): Dictionary of static settings that do not change during iterations.
    **kwargs: Additional keyword arguments for the solver.

  Returns:
    jnp.ndarray: Solution degrees of freedom after solving the linear system.

  Hessian Vector Product (HVP) Types:
    - 'fwdrev': Forward and reverse mode differentiation.
    - 'revrev': Reverse mode differentiation (only for symmetric matrices).
    - 'assemble': Assembles the tangent matrix explicitly (not supported with nodal imposition, then call e.g. PetSc).
    - 'linearize': Uses JAX's linearize function.
    - Default: Uses 'fwdrev' if not specified or if an unsupported type is provided.

  Solver Types:
    - 'cg': Conjugate Gradient.
    - 'normal cg': Normal Conjugate Gradient.
    - 'gmres': Generalized Minimal Residual Method.
    - 'bicgstab': BiConjugate Gradient Stabilized.
    - 'lu': LU Decomposition.
    - 'cholesky': Cholesky Decomposition (converts tangent to dense mode; not supported with nodal imposition).
    - 'qr': QR Decomposition (not supported with nodal imposition).
    - 'jacobi': Jacobi Method.
    - Default: Uses 'bicgstab' if not specified or if an unsupported type is provided.

  Notes:
    - Dirichlet boundary conditions are imposed if 'nodal imposition' is specified as the solution structure.
    - The itterative solver can be preconditioned with 'jacobi'.
    - When using 'assemble' HVP type, the function will explicitly assemble the tangent matrix.
  """
  solver = static_settings['solver']
  hvp_type = static_settings['hvp type']
  nodal_imposition = 'nodal imposition' in static_settings['solution structure']

  if nodal_imposition:
    # Impose Dirichlet boundaries. Dirichlet dofs has to be concrete, therefore it is passed in static_settings as tuple of tuples
    dirichlet_dofs_flat = np.asarray([element for tupl in static_settings['dirichlet dofs'] for element in tupl])
    free_dofs_flat = np.invert(dirichlet_dofs_flat)

    # Impose nodal dofs
    dirichlet_conditions = settings['dirichlet conditions'].flatten()
    flat_dofs = dofs.flatten()
    idx = jnp.arange(flat_dofs.shape[0])[dirichlet_dofs_flat]
    flat_dofs = flat_dofs.at[idx].set(dirichlet_conditions[idx])
    dofs = flat_dofs.reshape(dofs.shape)
    free_dofs = dofs.flatten()[free_dofs_flat]

    def residual_fun(x):
      flat_dofs = dofs.flatten()
      idx = jnp.arange(flat_dofs.shape[0])[free_dofs_flat]
      flat_dofs = flat_dofs.at[idx].set(x)
      return assembler.assemble_residual(flat_dofs.reshape(dofs.shape), settings, static_settings).flatten()[free_dofs_flat]

    def diag_assemble_fun(x):
      flat_dofs = dofs.flatten()
      idx = jnp.arange(flat_dofs.shape[0])[free_dofs_flat]
      flat_dofs = flat_dofs.at[idx].set(x)
      return assembler.assemble_tangent_diagonal(flat_dofs.reshape(dofs.shape), settings, static_settings).flatten()[free_dofs_flat]
    rhs = -residual_fun(dofs.flatten()[free_dofs_flat])

  else:
    free_dofs = dofs
    residual_fun = lambda x: assembler.assemble_residual(x, settings, static_settings)
    mat_assemble_fun = lambda x: assembler.assemble_tangent(x, settings, static_settings)
    diag_assemble_fun = lambda x: assembler.assemble_tangent_diagonal(x, settings, static_settings)
    rhs = -residual_fun(dofs)
 
  # Type of Hessian vector product
  match hvp_type:
    case 'fwdrev':
      def hvp_fwdrev(v):
        return jax.jvp(residual_fun, (free_dofs,), (v,))[1]
      hvp = hvp_fwdrev
    case 'revrev': # only works for symmetric matrices
      def hvp_revrev(v):
        return jax.jacrev(lambda x: jnp.vdot(residual_fun(x), v))(free_dofs)
      hvp = hvp_revrev
    case 'assemble':
      assert not nodal_imposition, "hvp type \'assemble\' is currently not supported for nodal imposition of DOFs."

      tangent = mat_assemble_fun(free_dofs)
      def hvp_assembled(v):
        mapped = tangent @ v.flatten()
        return jnp.reshape(mapped, v.shape)
      hvp = hvp_assembled
    case 'linearize':
      def hvp_linearized(v):
        (_,linearized) = jax.linearize(residual_fun, free_dofs)
        return linearized(v)
      hvp = hvp_linearized
    case _:
      hvp = hvp_fwdrev
      print('Type of hessian vector product has not been set or is not available. Using static_settings[\'matrix-free\'] = \'revrev\' as default.')

  # Preconditioning for itterative solvers
  if solver=='cg' or solver=='bicgstab' or solver=='normal cg' or solver=='gmres':
    try:
      precond_type = static_settings['type of preconditioner']
      match precond_type:
        case 'jacobi':
          # Inversion of diagonal part of tangent matrix as preconditioner
          M = 1 / diag_assemble_fun(free_dofs)
          def preconditioner(v):
            preconditioned = jnp.multiply(M, v.flatten())
            return jnp.reshape(preconditioned, free_dofs.shape)
        case _:
          def preconditioner(v):
            return v
          print("Wrong preconditioner keyword. Continue without preconditioner.")
    except KeyError:
      preconditioner = None
      pass

  # Select linear solver (itterative or direct)
  match solver:
    case 'cg':
      (sol, _)  = jax.scipy.sparse.linalg.cg(hvp, rhs, M=preconditioner, **kwargs)
    case 'normal cg':
      sol  = jaxopt.linear_solve.solve_normal_cg(hvp, rhs, **kwargs)
    case 'gmres':
      (sol, _)  = jax.scipy.sparse.linalg.gmres(hvp, rhs, M=preconditioner, **kwargs)
    case 'bicgstab':
      (sol, _)  = jax.scipy.sparse.linalg.bicgstab(hvp, rhs, M=preconditioner, **kwargs)
    case 'lu':
      sol  = jaxopt.linear_solve.solve_lu(hvp, rhs)
    case 'cholesky':
      assert not nodal_imposition, "solver \'cholesky\' is currently not supported for nodal imposition of DOFs."
      chol, lower = jax.scipy.linalg.cho_factor(mat_assemble_fun(dofs).todense())
      sol = jnp.reshape(jax.scipy.linalg.cho_solve((chol, lower), rhs.flatten()), dofs.shape)
    case 'qr':
      assert not nodal_imposition, "solver \'qr\' is currently not supported for nodal imposition of DOFs."
      bcoo_tangent = mat_assemble_fun(dofs)
      bcsr_tangent = sparse.BCSR.from_bcoo(bcoo_tangent).sum_duplicates(nse=bcoo_tangent.nse)
      sol = sparse.linalg.spsolve(bcsr_tangent.data, bcsr_tangent.indices, bcsr_tangent.indptr, rhs.flatten()).reshape(dofs.shape)
    case 'jacobi':
      diag = diag_assemble_fun(free_dofs)
      sol = jacobi_method(hvp, diag, free_dofs, rhs, **kwargs)
    case _:
      (sol, _)  = jax.scipy.sparse.linalg.bicgstab(hvp, rhs, M=preconditioner, **kwargs)
      print('Solver name not set or not available in combination with this solver type. Using static_settings[\'solver\'] = \'bicgstab\' as default.')

  if static_settings['verbose']>=1:
    residual = rhs - hvp(sol)
    jax.debug.print('The relative residual is: {value}', value = jnp.linalg.norm(residual) / jnp.linalg.norm(rhs))

  # Compose solution dofs
  if nodal_imposition:
    flat_dofs = dofs.flatten()
    flat_dofs = flat_dofs.at[idx].set(dirichlet_conditions[idx])
    idx = jnp.arange(flat_dofs.shape[0])[free_dofs_flat]
    flat_dofs = flat_dofs.at[idx].set(sol)
    dofs = flat_dofs.reshape(dofs.shape)
    return dofs
  else:
    return sol

def scipy_assembling(tangent_with_duplicates, verbose, free_dofs):
  """
  Convert a JAX BCOO matrix to a SciPy CSR matrix while summing duplicates.

  This function converts a JAX BCOO matrix, which may contain duplicate entries, 
  into a SciPy CSR matrix. It optionally deletes rows and columns corresponding 
  to Dirichlet degrees of freedom.

  Args:
    tangent_with_duplicates (jax.experimental.sparse.BCOO): The input JAX BCOO matrix.
    verbose (int): Verbosity level. If >= 2, timing information is printed.
    free_dofs (array or None): Boolean array indicating which degrees of freedom are free. 
                                If not None, rows and columns corresponding to Dirichlet DOFs 
                                are removed from the matrix.

  Returns:
    scipy.sparse.csr_matrix: The converted and possibly reduced CSR matrix.
  """
  if verbose>=2:
    start = time.time()

  data = tangent_with_duplicates.data
  indices = tangent_with_duplicates.indices
  rows = indices[:,0]
  cols = indices[:,1]

  # This is currently done on CPU and seems to be one of the computational bottlenecks when using GPU
  tangent_coo = scp.sparse.coo_matrix((data, (rows, cols)), shape=tangent_with_duplicates.shape)
  tangent_csr = scp.sparse.csr_matrix(tangent_coo)

  # Row deletion for Dirichlet-DOFs
  if free_dofs is not None:
    # Deleting rows and columns
    tangent_csr = tangent_csr[:, free_dofs]
    tangent_csr = tangent_csr[free_dofs]

  if verbose>=2:
    print("Time for summing duplicates / assemble tangent: ", time.time() - start)

  return tangent_csr

def linear_solve_petsc(mat, rhs, n_fields, solver, pc_type, verbose, free_dofs, tol=1e-8, **kwargs):
  """
  Solve a linear system using the PETSc solver (requires PETSc and petsc4py to be installed).

  This function solves a linear system using PETSc, with options for different 
  solvers and preconditioners. The input matrix is first converted to a SciPy 
  CSR matrix, and rows and columns corresponding to Dirichlet DOFs are optionally 
  removed.

  Args:
    mat (jax.experimental.sparse.BCOO): The input matrix in JAX BCOO format.
    rhs (jnp.ndarray): The right-hand side vector.
    n_fields (int): The number of fields.
    solver (str): The type of solver to use.
    pc_type (str): The type of preconditioner to use.
    verbose (int): Verbosity level. If >= 1, timing and solver information is printed.
    free_dofs (array or None): Boolean array indicating which degrees of freedom are free.
    tol (float): The relative tolerance for the solver.
    **kwargs: Additional keyword arguments for the solver.

  Returns:
    jax.numpy.array: The solution vector.
  
  Notes:
    - The solver settings can also be set from the command line. See PETSc and petsc4py documentation.
  """
  try:
    import petsc4py
    petsc4py.init(sys.argv)
    from petsc4py import PETSc
  except ModuleNotFoundError:
    print('Linear solver requires petsc and petsc4py')

  n_dofs = rhs.shape[0]
  
  # Transform matrix to csr format and sum duplicates
  tangent_csr = scipy_assembling(mat, verbose, free_dofs)
  if verbose>=2:
    start = time.time()

  # Load to petsc
  mat_petsc = PETSc.Mat().createAIJ(size=tangent_csr.shape, csr=(tangent_csr.indptr.astype(PETSc.IntType), tangent_csr.indices.astype(PETSc.IntType), tangent_csr.data))
  mat_petsc.setFromOptions()
  mat_petsc.setBlockSize(n_fields)
  
  if verbose>=2:
    print("To PETSc transformation time: ", time.time() - start)
    print("Matrix infos: ")
    print()
    print(mat_petsc.getInfo())
    start = time.time()

  # Initialization of right hand side and solution vector
  b = PETSc.Vec().createSeq(n_dofs)
  b.setFromOptions()
  b.setArray(np.array(rhs))
  x = PETSc.Vec().createSeq(n_dofs)
  x.setFromOptions()

  # Solver settings
  rtol = tol
  ksp = PETSc.KSP().create()
  ksp.setTolerances(rtol=rtol, **kwargs)
  ksp.setOperators(mat_petsc)
  ksp.setType(solver)
  ksp.setFromOptions()
  ksp.setConvergenceHistory()
  ksp.getPC().setType(pc_type)
  ksp.getPC().setFromOptions()
  
  # Monitoring
  if verbose>=2:
    print('Iteration   Residual')
    def monitor(ksp, its, rnorm):
      print('%5d      %20.15g'%(its,rnorm))
    ksp.setMonitor(monitor)

  # Solving
  ksp.solve(b, x)
  sol = jnp.asarray(x.getArray())

  if verbose>=2:
    residual = mat_petsc * x - b
    print("Itterative linear solver time: ", time.time() - start)
    print("Number of iterations: ", ksp.getIterationNumber())
    print("Type: ", ksp.getType())
    print("Tolerances: ", ksp.getTolerances())
    print(f"The relative residual is: {residual.norm() / b.norm()}.")
  return sol

def linear_solve_pardiso(mat, rhs, solver, verbose, free_dofs):
  """
  Solve a linear system using the PARDISO solver (requires Intel MKL and pypardiso('lu') or sparse_dot_mkl('qr') to be installed).

  This function solves a linear system using PARDISO, with options for different 
  solver types. The input matrix is first converted to a SciPy CSR matrix, and 
  rows and columns corresponding to Dirichlet DOFs are optionally removed.

  Args:
    mat (jax.experimental.sparse.BCOO): The input matrix in JAX BCOO format.
    rhs (jnp.ndarray): The right-hand side vector.
    solver (str): The type of solver to use ('lu' or 'qr').
    verbose (int): Verbosity level. If >= 2, timing information is printed.
    free_dofs (array or None): Boolean array indicating which degrees of freedom are free.

  Returns:
    jax.numpy.array: The solution vector.
  """
  # Transform matrix to csr format and sum duplicates
  tangent_csr = scipy_assembling(mat, verbose, free_dofs)
  if verbose>=2:
    start = time.time()

  # Prepare right hand side
  b = np.asarray(rhs)

  if solver == 'lu':
    try:
      import pypardiso
    except ModuleNotFoundError:
      print('Linear solver requires the installation of pypardiso.')

    # ToDo: make use of symmetries and other settings available#, msglvl=verbose, iparm=iparm
    # pypardiso_solver = pypardiso.PyPardisoSolver(mtype=11)
    # x = pypardiso.spsolve(tangent_csr, b, solver=pypardiso_solver)
    x = pypardiso.spsolve(tangent_csr, b)
  elif solver == 'qr':
    try:
      import sparse_dot_mkl
    except ModuleNotFoundError:
      print('Linear solver requires the installation of sparse_dot_mkl.')

    x = sparse_dot_mkl.sparse_qr_solve_mkl(tangent_csr, b)
  else:
      assert False, "Type of solver not supported. Choose \'lu\' or \'qr\'"

  sol = jnp.asarray(x)

  if verbose>=2:
    residual = b - tangent_csr * x
    print(f"The relative residual after linear solve is: {np.linalg.norm(residual) / np.linalg.norm(b)}.")
    print('Linear solver time: ', time.time() - start)
  return sol

def linear_solve_pyamg(mat, rhs, solver, pc_type, verbose, free_dofs, **kwargs):
  """
  Solve a linear system using the PyAMG solver (requires pyamg to be installed).

  This function solves a linear system using PyAMG, an algebraic multi-grid solver with options 
  for different solvers and preconditioners. The input matrix is first converted to a SciPy 
  CSR matrix, and rows and columns corresponding to Dirichlet DOFs are optionally 
  removed.

  Args:
    mat (jax.experimental.sparse.BCOO): The input matrix in JAX BCOO format.
    rhs (jnp.ndarray): The right-hand side vector.
    solver (str): The type of solver to use ('cg', 'bcgs', or 'gmres').
    pc_type (str): The type of preconditioner to use ('ruge stuben' or 'smoothed aggregation').
    verbose (int): Verbosity level. If >= 1, timing and solver information is printed.
    free_dofs (array or None): Boolean array indicating which degrees of freedom are free.
    **kwargs: Additional keyword arguments for the solver.

  Returns:
    jax.numpy.array: The solution vector.
  """
  # Transform matrix to csr format and sum duplicates
  pyamg_tangent = scipy_assembling(mat, verbose, free_dofs)

  if verbose>=2:
    start = time.time()

  # Set up solver
  try:
    import pyamg
  except ModuleNotFoundError:
    print('Linear solver requires the installation of pyamg.')

  if pc_type == 'ruge stuben':
    ml = pyamg.ruge_stuben_solver(A=pyamg_tangent)
  elif pc_type == 'smoothed aggregation':
    ml = pyamg.smoothed_aggregation_solver(A=pyamg_tangent)
  else:
      assert False, "Type of preconditioner not supported. Choose \'ruge stuben\' or \'smoothed aggregation\'"
  M = ml.aspreconditioner(cycle='V')

  if verbose>=2:
    print(ml)
    print("Time for setting up multigrid preconditioner: ", time.time() - start)
    start = time.time()

  # Prepare right hand side
  b = np.asarray(rhs)
  
  # Solving
  if solver == 'cg':
    x, _ = pyamg.krylov.cg(A=pyamg_tangent, b=b, M=M, **kwargs)
  elif solver == 'bcgs':
    x, _ = pyamg.krylov.bicgstab(A=pyamg_tangent, b=b, M=M, **kwargs)
  elif solver == 'gmres':
    x, _ = pyamg.krylov.gmres(A=pyamg_tangent, b=b, M=M, **kwargs)
  else:
      assert False, "Type of solver not supported. Choose \'cg\', \'bcgs\' or \'gmres\'"
  
  sol = jnp.asarray(x)

  if verbose>=2:
    residual = b - pyamg_tangent * x
    print(f"The relative residual is: {np.linalg.norm(residual) / np.linalg.norm(b)}.")
    print('Itterative linear solver time: ', time.time() - start)
  return sol

def linear_solve_scipy(mat, rhs, solver, verbose, free_dofs):
  """
  Solves a linear system using a specified SciPy solver.

  Parameters:
    mat (bcoo): JAX BCOO matrix representing the system's tangent matrix.
    rhs (jnp.ndarray): Right-hand side vector of the linear system.
    solver (str): Type of solver to use. Options are 'lapack' or 'umfpack'.
    verbose (int): Verbosity level for logging.
    free_dofs (jnp.ndarray): Boolean array indicating free degrees of freedom for Dirichlet boundary conditions.

  Returns:
    sol (jnp.ndarray): Solution vector to the linear system.
  """
  # Transform matrix to csr format and sum duplicates
  tangent_csr = scipy_assembling(mat, verbose, free_dofs)

  if verbose>=2:
    start = time.time()

  # Prepare right hand side
  b = rhs

  if solver == 'lapack':
    x = scp.sparse.linalg.spsolve(tangent_csr, b)
  elif solver == 'umfpack':
    x = scp.sparse.linalg.spsolve(tangent_csr, b, use_umfpack=True)
  else:
      assert False, "Type of solver not supported. Choose \'lapack\' or \'umfpack\'"
  
  sol = jnp.asarray(x)

  if verbose >=2:
    residual = b - tangent_csr * x
    print(f"The relative residual is: {np.linalg.norm(residual) / np.linalg.norm(b)}.")
    print('Direct solver time: ', time.time() - start)
  return sol


### Iterative solvers/smoothers
def jacobi_method(hvp_fun, diag, x_0, rhs, tol=1e-6, atol=1e-6, maxiter=1000):
  """
  Solve Ax = b using Jacobi iterations (experimental).

  Parameters:
    hvp_fun (function): Hessian vector product function.
    diag (jnp.ndarray): Diagonal of the Hessian matrix.
    x_0 (jnp.ndarray): Initial guess for the solution.
    rhs (jnp.ndarray): Right-hand side vector of the linear system.
    tol (float): Relative tolerance for convergence.
    atol (float): Absolute tolerance for convergence.
    maxiter (int): Maximum number of iterations.

  Returns:
    array: Solution vector x.
  """
  # Initialization
  inverse_diag = 1 / diag
  scaled_rhs = jnp.multiply(inverse_diag, rhs)
  rhs_squared = jnp.vdot(rhs, rhs)
  
  # Itterations
  def body_fun(value):
    x_k, k = value
    dx_k1 = scaled_rhs - jnp.multiply(inverse_diag, hvp_fun(x_k))
    x_k1 = x_k + dx_k1
    return (x_k1, k+1)

  def cond_fun(value):
    x_k, k = value
    r_k = rhs - hvp_fun(x_k)    
    r_k_squared = jnp.vdot(r_k, r_k)
    return k < maxiter #and r_k_squared > jnp.max(tol**2 * rhs_squared, atol**2)

  x_final, *_ = lax.while_loop(cond_fun, body_fun, (x_0, 0))

  return x_final

def damped_jacobi_relaxation(hvp_fun, diag, x_0, rhs, damping_factor=0.3333333, **kwargs):
  """
  Damped Jacobi smoother (experimental).

  Parameters:
    hvp_fun (callable): Hessian vector product function.
    diag (jnp.ndarray): Diagonal of the Hessian matrix.
    x_0 (jnp.ndarray): Initial guess for the solution.
    rhs (jnp.ndarray): Right-hand side vector of the linear system.
    damping_factor (float): Damping factor for the iterations (<=0.5 guarantees a good smoother).
    **kwargs (dict): Additional keyword arguments for customization.

  Returns:
    array: Solution vector x.
  """
  # Initialization
  inverse_diag = 1 / diag
  scaled_rhs = jnp.multiply(inverse_diag, rhs)
  
  # Itterations
  def body_fun(x_k, idx):
    dx_k1 = scaled_rhs - jnp.multiply(inverse_diag, hvp_fun(x_k))
    x_k1 = x_k + damping_factor * dx_k1
    return x_k1, None

  iterations = 20
  x_final, *_ = lax.scan(body_fun, init=x_0, xs=None, length=iterations)
  return x_final



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys
sys.path.insert(0, os.path.abspath('..'))

project = 'AutoPDEx'
copyright = '2024, Tobias Bode'
author = 'Tobias Bode'
release = '1.0.3'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration
extensions = [
    'sphinx.ext.napoleon',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.intersphinx',
    'sphinx.ext.viewcode',
    'sphinx.ext.mathjax',
    'sphinx_rtd_theme',
    'sphinx_copybutton',
    'sphinx_rtd_dark_mode',
    'nbsphinx',
    'jupyter_sphinx',
]

source_suffix = {
    '.rst': 'restructuredtext',
    '.md': 'markdown',
}

nbsphinx_allow_errors = True

mathjax_path = "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.7/MathJax.js?config=TeX-AMS-MML_HTMLorMML"

autosummary_generate = True
autodoc_default_flags = ['members', 'undoc-members', 'show-inheritance']

master_doc = 'index'

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output


import sphinx_rtd_theme
html_theme = 'sphinx_rtd_theme'
html_logo = '_static/logo.png'
html_favicon = ''
html_static_path = ['_static']

# Custom CSS
def setup(app):
    app.add_css_file('custom.css')

default_dark_mode = False


# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import time
import sys

import jax
from jax import lax, config
import jax.numpy as jnp
import flax
import pygmsh
import meshio

from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, spaces, assembler

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')


### Cook's membrane problem with simplex elements and two domains, linear elasticity
# First domain: domain integration
# Second domain: surface integration (tractions)


# h-refinement
post_values = []
for i in range(1,3):
  d_h = 10 / (2**i)

  ### Definition of geometry and boundary conditions
  pts = [[0.,0.], [48.,44.], [48.,60.], [0.,44.]]
  with pygmsh.geo.Geometry() as geom:
      geom.add_polygon(pts,mesh_size=d_h)
      mesh = geom.generate_mesh(order=2)

  # Mesh debugging
  mesh.write("mesh.vtk")

  # Import region mesh
  n_dim = 2
  x_nodes = jnp.asarray(mesh.points[:,:n_dim])
  n_nodes = x_nodes.shape[0]
  elements = jnp.asarray(mesh.cells_dict['triangle6'])
  surface_elements = jnp.asarray(mesh.cells_dict['line3'])

  # Selection of nodes for boundary condition
  dirichlet_nodes = geometry.in_planes(x_nodes, pts[0], [1.,0.])
  dirichlet_dofs = utility.dof_select(dirichlet_nodes, jnp.asarray([True, True]))
  dirichlet_conditions = jnp.zeros_like(dirichlet_dofs, dtype=jnp.float_)

  # Import surface mesh for inhomogeneous Neumann conditions
  neumann_selection_1 = geometry.select_elements_in_plane(x_nodes, surface_elements, pts[1], [1.,0.])
  neumann_elements_1 = surface_elements[neumann_selection_1]

  # Generate domain integration points in mesh
  (x_int, w_int, n_int, domain_connectivity) = seeder.int_pts_in_tri_mesh(x_nodes, elements, order=2)
  max_neighbors_1 = domain_connectivity.shape[-1]


  # Generate surface integration points on surface mesh and select domain elements that belong to the surface elements
  (x_surf_int, w_surf_int, n_surf_int, nods_of_surf_elem) = seeder.int_pts_in_line_mesh(x_nodes, neumann_elements_1, order=2)
  surf_connectivity = geometry.subelems_in_elems(nods_of_surf_elem, elements)
  max_neighbors_2 = surf_connectivity.shape[-1]

  # Definition of pde
  lam = 100.
  mu = 40.
  Em = mu * (3 * lam + 2 * mu) / (lam + mu)
  nu = lam / (2 * (lam + mu))
  youngs_mod_fun = lambda x: Em
  poisson_ratio_fun = lambda x: nu
  pde_1 = models.linear_elasticity_weak(youngs_mod_fun, poisson_ratio_fun, 'plain strain')

  # Tractions
  q_0 = 4.0
  traction_fun = lambda x: jnp.asarray([0., q_0])
  pde_2 = models.neumann_weak(traction_fun)


  ### Setting
  n_fields = 2
  dofs0 = jnp.zeros((n_nodes, n_fields))
  static_settings = flax.core.FrozenDict({
    'solution space': ('fem simplex', 'fem simplex'),
    'shape function mode': 'compiled',
    'number of fields': (n_fields, n_fields),
    'assembling mode': ('sparse', 'sparse'),
    'maximal number of neighbors': (max_neighbors_1, max_neighbors_2),
    'variational scheme': ('weak form galerkin', 'weak form galerkin'),
    'model': (pde_1, pde_2),
    'solution structure': ('nodal imposition', 'nodal imposition'),
    'solver type': 'newton',

    # 'solver backend': 'jax',
    # 'solver': 'cg',
    # 'hvp type': 'fwdrev',

    'solver backend': 'pardiso',
    'solver': 'lu',

    # 'solver backend': 'petsc',
    # 'solver': 'cg',
    # 'type of preconditioner': 'ilu',

    'verbose': 2,
    'dirichlet dofs': utility.jnp_to_tuple(dirichlet_dofs),
    'connectivity': (utility.jnp_to_tuple(domain_connectivity),utility.jnp_to_tuple(surf_connectivity)),
  })

  settings = {
    'node coordinates': x_nodes,
    'dirichlet conditions': dirichlet_conditions,
    'integration coordinates': (x_int, x_surf_int),
    'integration weights': (w_int, w_surf_int),
  }

  # Precompute shape functions
  settings = solution_structures.precompile(dofs0, settings, static_settings)


  # Call solver
  start = time.time()
  dofs, _ = solver.solver(dofs0, settings, static_settings)
  print("Analysis time: ", time.time() - start)
  print("Number of unknowns: ", dofs.flatten().shape)


  # Paraview postprocessing
  points = mesh.points
  cells = mesh.cells_dict["triangle6"]
  mesh = meshio.Mesh(
      points,
      {'triangle6': cells},
      point_data={
          "u": dofs,
      },
  )
  mesh.write("cooks_membrane.vtk")

  # Displacement at upper right edge
  idx = geometry.select_point(x_nodes, pts[2])
  v_ul = dofs[idx, 1].item()
  post_values.append([n_nodes * n_dim, v_ul]) # number of dofs over edge displacement in y-direction

print("Vertical displacements at upper right edge for each h-order:\n", post_values)



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

def test_implicit_diff():
    import time
    from functools import partial

    import jax
    from jax import config
    import jax.numpy as jnp
    import flax

    from autopdex import seeder, geometry, solver, utility, models, spaces

    config.update("jax_enable_x64", True)

    start = time.time()

    order = 2
    pts = [[0.,0.], [48.,44.], [48.,60.], [0.,44.]]
    n_dim = 2
    x_nodes = jnp.array([
        [ 0., 0.], [48., 44.], [48., 60.], [ 0., 44.], [12., 11.], [24., 22.], [36., 33.], [ 6., 5.5], [18., 16.5], 
        [30., 27.5], [42., 38.5], [48., 52.], [48., 48.], [48., 56.], [36., 56.], [24., 52.], [12., 48.], [42., 58.], 
        [30., 54.], [18., 50.], [ 6., 46.], [ 0., 33.], [ 0., 22.], [ 0., 11.], [ 0., 38.5], [ 0., 27.5], [ 0., 16.5], 
        [ 0., 5.5], [22.78099159, 39.33936105], [10.25815892, 28.78971905], [40.00000145, 47.99999935], [11.44621914, 36.55774821], 
        [30.81792875, 42.11576444], [ 7.62152819, 16.21744291], [23.3904958, 30.66968052], [17.11360537, 37.94855463], 
        [10.85218903, 32.67373363], [17.12907946, 25.39485952], [17.12134241, 31.67170708], [ 8.93984355, 22.50358098], 
        [ 9.81076409, 13.60872146], [13.46992178, 19.50179049], [38.00000072, 51.99999968], [44.00000072, 49.99999968], 
        [43.00000036, 53.99999984], [33.40896438, 37.55788222], [26.79946017, 40.72756274], [28.39973009, 34.11378137], 
        [38.00000072, 40.49999968], [43.00000036, 44.24999984], [11.72310957, 42.2788741], [23.3904958, 45.66968052], 
        [17.55680268, 43.97427731], [ 5.72310957, 34.7788741], [ 5.86155479, 40.38943705], [33.40896438, 49.05788222], 
        [35.70448255, 44.77894095], [ 5.12907946, 25.39485952], [ 5.42609452, 30.08686681], [28.39973009, 47.36378137], 
        [ 3.81076409, 13.60872146], [ 4.46992178, 19.50179049], [ 4.90538205, 9.55436073]
    ])
    n_nodes = x_nodes.shape[0]
    elements = jnp.array([
        [ 5, 28, 31, 29, 34, 35, 36, 37, 38], [ 4,  5, 29, 33,  8, 37, 39, 40, 41], [ 2, 14, 30, 11, 17, 42, 43, 13, 44],
        [ 5,  6, 32, 28,  9, 45, 46, 34, 47], [ 6,  1, 11, 30, 10, 12, 43, 48, 49], [15, 16, 31, 28, 19, 50, 35, 51, 52], 
        [16,  3, 21, 31, 20, 24, 53, 50, 54], [ 6, 30, 14, 32, 48, 42, 55, 45, 56], [22, 29, 31, 21, 57, 36, 53, 25, 58], 
        [14, 15, 28, 32, 18, 51, 46, 55, 59], [23, 33, 29, 22, 60, 39, 57, 26, 61], [33, 23,  0,  4, 60, 27,  7, 40, 62]
    ])
    surface_elements = jnp.array([
        [ 0,  4,  7], [ 4,  5,  8], [ 5,  6,  9], [ 6,  1, 10], [ 1, 11, 12], [11,  2, 13], [ 2, 14, 17], 
        [14, 15, 18], [15, 16, 19], [16,  3, 20], [ 3, 21, 24], [21, 22, 25], [22, 23, 26], [23,  0, 27]
    ])
    # import sys
    # sys.exit()

    # Selection of nodes for boundary condition
    dirichlet_nodes = geometry.in_planes(x_nodes, pts[0], [1.,0.])
    dirichlet_dofs = utility.dof_select(dirichlet_nodes, jnp.asarray([True, True]))
    dirichlet_conditions = jnp.zeros_like(dirichlet_dofs, dtype=jnp.float64)

    # Import surface mesh for inhomogeneous Neumann conditions
    neumann_selection = geometry.select_elements_in_plane(x_nodes, surface_elements, pts[1], [1.,0.])
    neumann_elements= surface_elements[neumann_selection]


    ### Built-in autopdex element
    weak_form_fun_1 = models.hyperelastic_steady_state_weak(
    models.neo_hooke, 
    lambda x, settings: settings['youngs modulus'],
    lambda x, settings: settings['poisson ratio'], 
    'plain strain')
    user_elem_1 = models.isoparametric_domain_element_galerkin(
    weak_form_fun_1, 
    spaces.fem_iso_line_quad_brick,
    *seeder.gauss_legendre_nd(dimension = 2, order = 2 * order))

    # Tractions
    q_0 = 2.0e+1
    traction_fun = lambda x, settings: jnp.asarray([0., settings['load multiplier']])
    weak_form_fun_2 = models.neumann_weak(traction_fun)
    user_elem_2 = models.isoparametric_surface_element_galerkin(
    weak_form_fun_2, 
    spaces.fem_iso_line_quad_brick,
    *seeder.gauss_legendre_nd(dimension = 1, order = 2 * order),
    tangent_contributions=False)

    # Adaptive load stepping
    def multiplier_settings(settings, multiplier):
        settings['load multiplier'] = multiplier * q_0
        return settings

    ### Settings
    n_fields = 2
    dofs_0 = jnp.zeros((n_nodes, n_fields))
    static_settings = flax.core.FrozenDict({
    'number of fields': (n_fields, n_fields),
    'assembling mode': ('user element', 'user element'),
    'solution structure': ('nodal imposition', 'nodal imposition'),
    'model': (user_elem_1, user_elem_2),
    'solver type': 'newton',
    'solver backend': 'scipy',
    'solver': 'lapack',
    'verbose': -1,
    'dirichlet dofs': utility.jnp_to_tuple(dirichlet_dofs),
    'connectivity': (utility.jnp_to_tuple(elements), utility.jnp_to_tuple(neumann_elements)),
    })
    settings = {
    'load multiplier': q_0,
    'node coordinates': x_nodes,
    'dirichlet conditions': dirichlet_conditions,
    }

    youngs_mod_mean = 100.
    nu_mean = 0.3

    # Compare backward and forward mode implicit diff with precomputed values
    @partial(jax.jit, static_argnames=['static_settings'])
    def test_results(dofs_0, settings, static_settings):
        # Second order mixed sensitivities, forward mode
        def fun(dofs, settings, Em, nu):
            settings['youngs modulus'] = Em
            settings['poisson ratio'] = nu
            dofs = solver.adaptive_load_stepping(dofs, settings, static_settings, multiplier_settings, False, 'forward', newton_tol=1e-8)[0]
            return dofs.flatten() @ dofs.flatten()
        forward_and_sensitivity = lambda v: fun(dofs_0, settings, v[0], v[1])
        result = utility.jacfwd_upto_n_one_vector_arg(forward_and_sensitivity, jnp.asarray([youngs_mod_mean, nu_mean]), 2)
        dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dEmdnu, d2u_dnudEm, d2u_dnu2 = result[0], result[1][0], result[1][1], result[2][0][0], result[2][0][1], result[2][1][0], result[2][1][1]

        # Reverse mode implicit diff:
        def fun_bwd(dofs, settings, Em, nu):
            settings['youngs modulus'] = Em
            settings['poisson ratio'] = nu
            dofs = solver.adaptive_load_stepping(dofs, settings, static_settings, multiplier_settings, False, 'backward', newton_tol=1e-8)[0]
            return dofs.flatten() @ dofs.flatten()
        def compute_sensitivities_backward(v):
            sol = fun_bwd(dofs_0, settings, v[0], v[1])
            du_dnu = jax.jacrev(fun_bwd, argnums=3)(dofs_0, settings, v[0], v[1])
            d2u_dnu2 = jax.jacrev(jax.jacrev(fun_bwd, argnums=3), argnums=3)(dofs_0, settings, v[0], v[1])
            return sol, du_dnu, d2u_dnu2
        sol_rev, dudnu_rev, d2udnu2_rev = compute_sensitivities_backward(jnp.asarray([youngs_mod_mean, nu_mean]))

        return dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dEmdnu, d2u_dnudEm, d2u_dnu2, sol_rev, dudnu_rev, d2udnu2_rev
    dofs, du_dEm, du_dnu, d2u_dEm2, d2u_dEmdnu, d2u_dnudEm, d2u_dnu2, sol_rev, dudnu_rev, d2udnu2_rev = test_results(dofs_0, settings, static_settings)


        # Precomputed values:
    test1, test2, test3, test4, test5, test6, test7 = (
        jnp.array(19390.35027108, dtype=jnp.float64), 
        jnp.array(-216.0310416, dtype=jnp.float64), 
        jnp.array(-1859.43760286, dtype=jnp.float64), 
        jnp.array(4.17157824, dtype=jnp.float64), 
        jnp.array(-51.42695148, dtype=jnp.float64), 
        jnp.array(-51.42695148, dtype=jnp.float64), 
        jnp.array(-45854.7002203, dtype=jnp.float64))

    # Tests
    assert jnp.allclose(dofs, test1), 'Incorrect solution.'
    assert jnp.allclose(du_dEm, test2), 'Incorrect first order sensitivities in forward mode'
    assert jnp.allclose(du_dnu, test3), 'Incorrect first order sensitivities in forward mode'
    assert jnp.allclose(d2u_dEm2, test4), 'Incorrect second order sensitivities in forward mode'
    assert jnp.allclose(d2u_dEmdnu, test5), 'Incorrect second order sensitivities in forward mode'
    assert jnp.allclose(d2u_dnudEm, test6), 'Incorrect second order sensitivities in forward mode'
    assert jnp.allclose(d2u_dnu2, test7), 'Incorrect second order sensitivities in forward mode'

    assert jnp.allclose(sol_rev, test1), 'Incorrect solution.'
    assert jnp.allclose(dudnu_rev, test3), 'Incorrect in first order sensitivity in backward mode.'
    assert jnp.allclose(d2udnu2_rev, test7), 'Incorrect in first order sensitivity in backward mode.'

    # computing_time = time.time() - start
    # print('All tests passed. Compute time: ', computing_time)




# assembler.py
# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
Module for assembling and integrating functionals, residuals and tangents, supporting both 'sparse' and 'dense' modes and 'user potentials', 'user residuals' and 'user elements'.

The assembly modes 'dense' and 'sparse' use JAX's automatic differentiation capabilities to determine the global residual and the tangent for the given values of the degrees of freedom and the models and variational schemes specified in static_settings.
In the 'dense' mode, the tangent matrix is returned as a jnp.ndarray, while in the 'sparse' mode, it is returned as a jax.experimental.BCOO matrix with duplicates. 
In the 'user potential', 'user residual' and 'user element' modes, the global residual and tangent can be assembled based on user-defined element-wise contributions. 
Depending on the execution location, the entries are calculated on CPU or GPU. 
For an efficient calculation, JAX's automatic vectorization transformation (vmap) is used.
The summation of duplicates is then carried out within the solver module. Currently, SciPy is used for this on the CPU.
"""

import jax
import jax.numpy as jnp
from jax.experimental import sparse
import sparsejac
import numpy as np

from autopdex import variational_schemes
from autopdex.utility import jit_with_docstring


### Assembling switches: sparse/dense
@jit_with_docstring(static_argnames=['static_settings'])
def integrate_functional(dofs, settings, static_settings):
  """
  Integrate functional as sum over set of domains.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
  
  Returns:
    float: The integrated functional value for current dofs.
  """
  # ToDo: add a function that checks the entries of settings and static_settings for inconsistencies

  # Loop over all sets of integration points/ domains
  num_sets = len(static_settings['assembling mode'])
  integrated_functional = 0
  for set in range(num_sets):
    assembling_mode = static_settings['assembling mode'][set]

    if assembling_mode == 'dense':
      integrated_functional += dense_integrate_functional(dofs, settings, static_settings, set)
    elif assembling_mode == 'sparse':
      integrated_functional += sparse_integrate_functional(dofs, settings, static_settings, set)
    elif assembling_mode == 'user potential':
      integrated_functional += user_potential_integrate_functional(dofs, settings, static_settings, set)
    else:
      assert False, 'Assembling mode can be either \'sparse\' or \'dense\' in integrate_functional'

  return integrated_functional

@jit_with_docstring(static_argnames=['static_settings'])
def assemble_residual(dofs, settings, static_settings):
  """
  Assemble residuals over set of domains.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
  
  Returns:
    jnp.ndarray: The assembled residual.
  """
  # Loop over all sets of integration points/ domains
  num_sets = len(static_settings['assembling mode'])
  integrated_residual = jnp.zeros_like(dofs)
  for set in range(num_sets):
    assembling_mode = static_settings['assembling mode'][set]

    if assembling_mode == 'dense':
      integrated_residual += dense_assemble_residual(dofs, settings, static_settings, set)
    elif assembling_mode == 'sparse':
      integrated_residual += sparse_assemble_residual(dofs, settings, static_settings, set)
    elif assembling_mode == 'user potential':
      integrated_residual += user_potential_assemble_residual(dofs, settings, static_settings, set)
    elif assembling_mode == 'user residual':
      integrated_residual += user_residual_assemble_residual(dofs, settings, static_settings, set)
    elif assembling_mode == 'user element':
      integrated_residual += user_element_assemble_residual(dofs, settings, static_settings, set)
    else:
      assert False, 'Assembling mode can be either \'sparse\', \'dense\' or \'user element\''

  return integrated_residual

@jit_with_docstring(static_argnames=['static_settings'])
def assemble_tangent_diagonal(dofs, settings, static_settings):
  """
  Assemble the diagonal of the tangent matrix.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
  
  Returns:
    jnp.ndarray: The diagonal of the assembled tangent matrix.
  """
  # Loop over all sets of integration points/ domains
  num_sets = len(static_settings['assembling mode'])
  tangent_diagonal = jnp.zeros_like(dofs.flatten())
  for set in range(num_sets):
    assembling_mode = static_settings['assembling mode'][set]
    if assembling_mode == 'sparse':
      tangent_diagonal += sparse_assemble_tangent_diagonal(dofs, settings, static_settings, set)
    elif assembling_mode == 'user potential':
      tangent_diagonal += user_potential_assemble_tangent_diagonal(dofs, settings, static_settings, set)
    elif assembling_mode == 'user residual':
      tangent_diagonal += user_residual_assemble_tangent_diagonal(dofs, settings, static_settings, set)
    elif assembling_mode == 'user element':
      tangent_diagonal += user_element_assemble_tangent_diagonal(dofs, settings, static_settings, set)
    else:
      assert False, 'Assembling mode for assembling tangent diagonal supports currently only \'sparse\' and \'user element\''
  return tangent_diagonal

@jit_with_docstring(static_argnames=['static_settings'])
def assemble_tangent(dofs, settings, static_settings):
  """
  Assemble the full (possibly sparse) tangent matrix.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
  
  Returns:
    jnp.ndarray or sparse matrix: The assembled tangent matrix.
  """
  # ToDo: add symmetric mode

  num_sets = len(static_settings['assembling mode'])
  one_dense = 'dense' in static_settings['assembling mode']
  num_dofs = np.prod(dofs.shape)
  try:
    sparsity_pattern = static_settings['known sparsity pattern']
  except KeyError:
    sparsity_pattern = 'none'

  match sparsity_pattern:
    case 'none':
      if one_dense:
        integrated_tangent = jnp.zeros((num_dofs, num_dofs))
      else:
        integrated_tangent = sparse.empty((num_dofs, num_dofs), dtype=jnp.float_, index_dtype=jnp.int_)

      # Loop over all sets of integration points/ domains
      for set in range(num_sets):
        assembling_mode = static_settings['assembling mode'][set]

        if assembling_mode == 'dense':
          integrated_tangent += dense_assemble_tangent(dofs, settings, static_settings, set)
        else:
          if assembling_mode == 'sparse':
            add = sparse_assemble_tangent(dofs, settings, static_settings, set)
          elif assembling_mode == 'user potential':
            add = user_potential_assemble_tangent(dofs, settings, static_settings, set)
          elif assembling_mode == 'user residual':
            add = user_residual_assemble_tangent(dofs, settings, static_settings, set)
          elif assembling_mode == 'user element':
            add = user_element_assemble_tangent(dofs, settings, static_settings, set)
          else:
            assert False, 'Assembling mode can be either \'sparse\', \'dense\' or \'user element\''

          if one_dense:
            integrated_tangent += add.todense()
          else:
            integrated_tangent += add
    case 'diagonal':
      # Compute the diagonal tangent (not the diagonal of a tangent that is not diagonal)
      residual_fun = lambda flat_dofs: assemble_residual(flat_dofs.reshape(dofs.shape), settings, static_settings).flatten()
      with jax.ensure_compile_time_eval():
        data_and_indices = (jnp.ones((num_dofs,)),
                            jax.vmap(lambda i: jnp.asarray([i, i]))(jnp.arange(0, num_dofs)))
        mat_shape = (num_dofs,num_dofs)
        sparsity = jax.experimental.sparse.BCOO(data_and_indices, shape=mat_shape) # ToDo: with unique and sorted indices more efficient?
        sparse_diag_fun = sparsejac.jacfwd(residual_fun, sparsity=sparsity)
      diag = sparse_diag_fun(dofs.flatten())
      return diag

    case _:
      assert False, '\'known sparsity pattern\' mode is not implemented.'

  return integrated_tangent


### Dense assembling
@jit_with_docstring(static_argnames=['static_settings', 'set'])
def dense_integrate_functional(dofs, settings, static_settings, set):
  """
  Dense integration of functional of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    float: The integrated functional value.
  """
  x_int = settings['integration coordinates'][set]
  w_int = settings['integration weights'][set]

  functional_at_int_point_vj = jax.vmap(variational_schemes.functional_at_int_point, (0,0,0,None,None,None,None), 0)
  int_point_numbers = jnp.arange(0, x_int.shape[0], 1)
  integrated_functional = functional_at_int_point_vj(x_int, w_int, int_point_numbers, dofs, settings, static_settings, set).sum()

  return integrated_functional

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def dense_assemble_residual(dofs, settings, static_settings, set):
  """
  Dense assembly of residual of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jnp.ndarray: The assembled residual.
  """
  grad = jax.grad(dense_integrate_functional, set)
  return grad(dofs, settings, static_settings).flatten()

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def dense_assemble_tangent(dofs, settings, static_settings, set):
  """
  Dense assembly of tangent of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jnp.ndarray: The assembled tangent matrix.
  """
  hessian = jax.hessian(dense_integrate_functional)
  return jnp.reshape(hessian(dofs, settings, static_settings, set),
                      (dofs.shape[0]*dofs.shape[1],dofs.shape[0]*dofs.shape[1]))


### Sparse assembling
@jit_with_docstring(static_argnames=['static_settings', 'set'])
def sparse_integrate_functional(dofs, settings, static_settings, set):
  """
  Sparse integration of functional of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    float: The integrated functional value.
  """
  neighbor_list = jnp.asarray(static_settings['connectivity'][set])
  local_dofs = dofs[neighbor_list]
  x_int = settings['integration coordinates'][set]
  w_int = settings['integration weights'][set]

  functional_at_int_point_vj = jax.vmap(variational_schemes.functional_at_int_point, (0,0,0,0,None,None,None), 0)
  int_point_numbers = jnp.arange(0, x_int.shape[0], 1)
  return functional_at_int_point_vj(x_int, w_int, int_point_numbers, local_dofs, settings, static_settings, set).sum()

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def sparse_assemble_residual(dofs, settings, static_settings, set):
  """
  Sparse assembly of residual of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jnp.ndarray: The assembled residual.
  """
  neighbor_list = jnp.asarray(static_settings['connectivity'][set])
  local_dofs = dofs[neighbor_list]
  variational_scheme = static_settings['variational scheme'][set]
  n_fields = static_settings['number of fields'][set]
  x_int = settings['integration coordinates'][set]
  w_int = settings['integration weights'][set]

  int_point_numbers = jnp.arange(0, x_int.shape[0], 1)

  if variational_scheme == 'least square pde loss' or variational_scheme == 'least square function approximation':
    grad = jax.jacrev(sparse_integrate_functional)
    return grad(dofs, settings, static_settings, set)

  elif variational_scheme == 'strong form galerkin':
    # Direct implementation of residual, e.g. for Galerkin method
    residual_at_int_point_vj = jax.vmap(variational_schemes.direct_residual_at_int_point, (0,0,0,0,None,None,None), 0)
    residual_contributions = residual_at_int_point_vj(x_int, w_int, int_point_numbers, local_dofs, settings, static_settings, set)

    flat_neighbors = neighbor_list.flatten()
    tmp = jnp.outer(jnp.ones(flat_neighbors.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_))
    indices = tmp + jnp.outer(flat_neighbors, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)

    res = jnp.zeros(dofs.shape).flatten()
    res = res.at[indices.flatten()].add(residual_contributions.flatten())
    res = res.reshape(dofs.shape)
  elif variational_scheme == 'weak form galerkin':
    # Pass local_dofs twice (assuming Bubnov Galerkin...)
    residual_at_int_point_vj = jax.vmap(variational_schemes.residual_from_deriv_at_int_point, (0,0,0,0,0,None,None,None), 0)
    residual_contributions = residual_at_int_point_vj(x_int, w_int, int_point_numbers, local_dofs, local_dofs, settings, static_settings, set)

    # Remaining stuff like for strong form galerkin
    flat_neighbors = neighbor_list.flatten()
    tmp = jnp.outer(jnp.ones(flat_neighbors.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_))
    indices = tmp + jnp.outer(flat_neighbors, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)

    res = jnp.zeros(dofs.shape).flatten()
    res = res.at[indices.flatten()].add(residual_contributions.flatten())
    res = res.reshape(dofs.shape)
    res = jnp.reshape(res, dofs.shape)
  else:
    assert False, 'Residual mode not or wrongly specified!'

  # Assembled residual
  return res

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def sparse_assemble_tangent_diagonal(dofs, settings, static_settings, set):
  """
  Sparse assembly of the diagonal of the tangent matrix for specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jnp.ndarray: The diagonal of the assembled tangent matrix.
  """

  neighbor_list = jnp.asarray(static_settings['connectivity'][set])
  local_dofs = dofs[neighbor_list]
  maximum = static_settings['maximal number of neighbors'][set]
  n_fields = static_settings['number of fields'][set]
  variational_scheme = static_settings['variational scheme'][set]
  x_int = settings['integration coordinates'][set]
  w_int = settings['integration weights'][set]

  # Compute tangent contributions
  if variational_scheme == 'least square pde loss' or variational_scheme == 'least square function approximation':
    tangent_at_int_point = jax.jacfwd(jax.jacrev(variational_schemes.functional_at_int_point, argnums=3), argnums=3) # argnum=3 are unknowns with local support
  elif variational_scheme == 'strong form galerkin':
    tangent_at_int_point = jax.jacfwd(variational_schemes.direct_residual_at_int_point, argnums=3)
  elif variational_scheme == 'weak form galerkin':
    tangent_at_int_point = jax.jacfwd(variational_schemes.residual_from_deriv_at_int_point, argnums=3)
  else:
    assert False, 'Residual mode not or wrongly specified!'

  int_point_numbers = jnp.arange(0, x_int.shape[0], 1)

  # Same as for assembling the whole matrix, but just take the diagonal on the local level
  if variational_scheme == 'weak form galerkin':
    tangent_at_int_point_vj = jax.vmap(tangent_at_int_point, (0,0,0,0,0,None,None,None), 0)
    tangent_contributions = tangent_at_int_point_vj(x_int, w_int, int_point_numbers, local_dofs, local_dofs, settings, static_settings, set)
  else:
    tangent_at_int_point_vj = jax.vmap(tangent_at_int_point, (0,0,0,0,None,None,None), 0)
    tangent_contributions = tangent_at_int_point_vj(x_int, w_int, int_point_numbers, local_dofs, settings, static_settings, set)
  tangent_contributions = jnp.reshape(tangent_contributions, (x_int.shape[0], maximum*n_fields, maximum*n_fields))
  diagonal_contributions = jnp.diagonal(tangent_contributions, axis1=1, axis2=2)

  # Assemble diagonal entries like a residual
  flat_neighbors = neighbor_list.flatten()
  tmp = jnp.outer(jnp.ones(flat_neighbors.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_))
  indices = tmp + jnp.outer(flat_neighbors, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)
  diag = jnp.zeros(dofs.shape).flatten()
  diag = diag.at[indices.flatten()].add(diagonal_contributions.flatten())
  return diag

# ToDo: Try row-wise assembling with automatic treatment of duplicates (maybe with sparsejac and pre-calculation of sparsity pattern with ints)
@jit_with_docstring(static_argnames=['static_settings', 'set'])
def sparse_assemble_tangent(dofs, settings, static_settings, set):
  """
  Sparse assembly of the full tangent matrix of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jax.experimental.sparse.BCOO: The assembled tangent matrix.
  """
  neighbor_list = jnp.asarray(static_settings['connectivity'][set])
  local_dofs = dofs[neighbor_list]
  maximum = static_settings['maximal number of neighbors'][set]
  n_fields = static_settings['number of fields'][set]
  variational_scheme = static_settings['variational scheme'][set]
  x_int = settings['integration coordinates'][set]
  w_int = settings['integration weights'][set]

  # Compute tangent contributions
  if variational_scheme == 'least square pde loss' or variational_scheme == 'least square function approximation':
    tangent_at_int_point = jax.jacfwd(jax.jacrev(variational_schemes.functional_at_int_point, argnums=3), argnums=3) # argnum=3 are unknowns with local support
  elif variational_scheme == 'strong form galerkin':
    tangent_at_int_point = jax.jacfwd(variational_schemes.direct_residual_at_int_point, argnums=3)
  elif variational_scheme == 'weak form galerkin':
    tangent_at_int_point = jax.jacfwd(variational_schemes.residual_from_deriv_at_int_point, argnums=3)
  else:
    assert False, 'Residual mode not or wrongly specified!'
    
  int_point_numbers = jnp.arange(0, x_int.shape[0], 1)
  if variational_scheme == 'weak form galerkin':
    tangent_at_int_point_vj = jax.vmap(tangent_at_int_point, (0,0,0,0,0,None,None,None), 0)
    tangent_contributions = tangent_at_int_point_vj(x_int, w_int, int_point_numbers, local_dofs, local_dofs, settings, static_settings, set)
  else:
    tangent_at_int_point_vj = jax.vmap(tangent_at_int_point, (0,0,0,0,None,None,None), 0)
    tangent_contributions = tangent_at_int_point_vj(x_int, w_int, int_point_numbers, local_dofs, settings, static_settings, set)
  tangent_contributions = jnp.reshape(tangent_contributions, (x_int.shape[0], maximum*n_fields, maximum*n_fields))

  # Assembling (without summing duplicates)
  def get_indices(list):
    def one_elem_idx(neighb):
      def row_indices(a, row):
        return jnp.asarray([[a,row[i]] for i in range(row.shape[0])])
      colrow = jax.vmap(row_indices, (0, None), 0)
      tmp = jnp.outer(jnp.ones(neighb.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_)).astype(jnp.int_)
      indices = tmp + jnp.outer(neighb, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)
      flat_indices = indices.flatten()
      return colrow(flat_indices,flat_indices)
    all_elem_idx = jax.vmap(one_elem_idx, (0), 0)
    return jnp.reshape(all_elem_idx(list), shape=(list.shape[0]*(maximum*n_fields)**2,2))
  indices = get_indices(neighbor_list)
  data = tangent_contributions.flatten()
  n_dofs = n_fields * dofs.shape[0]

  tangent = sparse.BCOO((data, indices), shape=(n_dofs,n_dofs))
  return tangent


### Assembling for user potentials
@jit_with_docstring(static_argnames=['static_settings', 'set'])
def user_potential_integrate_functional(dofs, settings, static_settings, set):
  """
  Assembly of potential for custom user definition of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    float: value of functional integrated over set of elements
  """

  user_elem_fun = static_settings['model'][set]
  connectivity_list = jnp.asarray(static_settings['connectivity'][set])
  x_nodes = settings['node coordinates']
  local_dofs = dofs[connectivity_list]
  local_node_coor = x_nodes[connectivity_list]
  elem_numbers = jnp.arange(0, connectivity_list.shape[0], 1)

  # For each element call user element routine (as the one in pde-module with mode='residual') and get contributions
  functional_contributions = jax.vmap(user_elem_fun, (0, 0, 0, None, None, None), (0)
                                      )(local_dofs, local_node_coor, elem_numbers, settings, static_settings, set)

  return functional_contributions.sum()

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def user_potential_assemble_residual(dofs, settings, static_settings, set):
  """
  Assembly of residual for custom user potential of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jnp.ndarray: The assembled residual.
  """
  n_fields = static_settings['number of fields'][set]
  user_potential_fun = static_settings['model'][set]
  connectivity_list = jnp.asarray(static_settings['connectivity'][set])
  x_nodes = settings['node coordinates']
  local_dofs = dofs[connectivity_list]
  local_node_coor = x_nodes[connectivity_list]
  elem_numbers = jnp.arange(0, connectivity_list.shape[0], 1)

  # For each element call user residual routine (as the one in pde-module with mode='residual') and get contributions
  residual_contributions = jax.vmap(jax.jacrev(user_potential_fun), (0, 0, 0, None, None, None), (0))(local_dofs, local_node_coor, elem_numbers, settings, static_settings, set)

  # Assembling
  flat_neighbors = connectivity_list.flatten()
  tmp = jnp.outer(jnp.ones(flat_neighbors.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_))
  indices = tmp + jnp.outer(flat_neighbors, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)

  res = jnp.zeros(dofs.shape).flatten()
  res = res.at[indices.flatten()].add(residual_contributions.flatten())
  res = res.reshape(dofs.shape)
  res = jnp.reshape(res, dofs.shape)
  return res

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def user_potential_assemble_tangent_diagonal(dofs, settings, static_settings, set):
  """
  Assembly of the diagonal of the tangent matrix for custom user potential of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jnp.ndarray: The diagonal of the assembled tangent matrix.
  """
  n_fields = static_settings['number of fields'][set]
  user_potential_fun = static_settings['model'][set]
  connectivity_list = jnp.asarray(static_settings['connectivity'][set])
  x_nodes = settings['node coordinates']
  local_dofs = dofs[connectivity_list]
  local_node_coor = x_nodes[connectivity_list]
  elem_numbers = jnp.arange(0, connectivity_list.shape[0], 1)

  # For each element call user element routine (as the one in pde-module with mode='residual') and get contributions
  tangent_contributions = jax.vmap(jax.jacfwd(jax.jacrev(user_potential_fun)), (0, 0, 0, None, None, None), (0))(local_dofs, local_node_coor, elem_numbers, settings, static_settings, set)
  diagonal_contributions = jnp.diagonal(tangent_contributions, axis1=1, axis2=2)

  # Assemble diagonal entries like a residual
  flat_neighbors = connectivity_list.flatten()
  tmp = jnp.outer(jnp.ones(flat_neighbors.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_))
  indices = tmp + jnp.outer(flat_neighbors, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)
  diag = jnp.zeros(dofs.shape).flatten()
  diag = diag.at[indices.flatten()].add(diagonal_contributions.flatten())
  return diag

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def user_potential_assemble_tangent(dofs, settings, static_settings, set):
  """
  Assembly of the full (sparse) tangent matrix for custom user potential of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jax.experimental.sparse.BCOO: The assembled tangent matrix.
  """
  n_fields = static_settings['number of fields'][set]
  user_potential_fun = static_settings['model'][set]
  connectivity_list = jnp.asarray(static_settings['connectivity'][set])
  x_nodes = settings['node coordinates']
  local_dofs = dofs[connectivity_list]
  local_node_coor = x_nodes[connectivity_list]
  dofs_per_elem = local_dofs[0].flatten().shape[0]
  elem_numbers = jnp.arange(0, connectivity_list.shape[0], 1)

  # For each element call user element routine (as the one in pde-module with mode='residual') and get contributions
  tangent_contributions = jax.vmap(jax.jacfwd(jax.jacrev(user_potential_fun)), (0, 0, 0, None, None, None), (0)
                                   )(local_dofs, local_node_coor, elem_numbers, settings, static_settings, set)

  # Assembling (without summing duplicates)
  def get_indices(list):
    def one_elem_idx(neighb):
      def row_indices(a, row):
        return jnp.asarray([[a,row[i]] for i in range(row.shape[0])])
      colrow = jax.vmap(row_indices, (0, None), 0)
      tmp = jnp.outer(jnp.ones(neighb.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_)).astype(jnp.int_)
      indices = tmp + jnp.outer(neighb, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)
      flat_indices = indices.flatten()
      return colrow(flat_indices,flat_indices)
    all_elem_idx = jax.vmap(one_elem_idx, (0), 0)
    return jnp.reshape(all_elem_idx(list), shape=(list.shape[0]*(dofs_per_elem)**2,2))
  
  indices = get_indices(connectivity_list)
  data = tangent_contributions.flatten()
  n_dofs = n_fields * dofs.shape[0]

  tangent = sparse.BCOO((data, indices), shape=(n_dofs,n_dofs))
  return tangent


### Assembling for user residuals
@jit_with_docstring(static_argnames=['static_settings', 'set'])
def user_residual_assemble_residual(dofs, settings, static_settings, set):
  """
  Assembly of residual for custom user residual of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jnp.ndarray: The assembled residual.
  """
  n_fields = static_settings['number of fields'][set]
  user_residual_fun = static_settings['model'][set]
  connectivity_list = jnp.asarray(static_settings['connectivity'][set])
  x_nodes = settings['node coordinates']
  local_dofs = dofs[connectivity_list]
  local_node_coor = x_nodes[connectivity_list]
  elem_numbers = jnp.arange(0, connectivity_list.shape[0], 1)

  # For each element call user residual routine (as the one in pde-module with mode='residual') and get contributions
  residual_contributions = jax.vmap(user_residual_fun, (0, 0, 0, None, None, None), (0))(local_dofs, local_node_coor, elem_numbers, settings, static_settings, set)

  # Assembling
  flat_neighbors = connectivity_list.flatten()
  tmp = jnp.outer(jnp.ones(flat_neighbors.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_))
  indices = tmp + jnp.outer(flat_neighbors, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)

  res = jnp.zeros(dofs.shape).flatten()
  res = res.at[indices.flatten()].add(residual_contributions.flatten())
  res = res.reshape(dofs.shape)
  res = jnp.reshape(res, dofs.shape)
  return res

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def user_residual_assemble_tangent_diagonal(dofs, settings, static_settings, set):
  """
  Assembly of the diagonal of the tangent matrix for custom user residual of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jnp.ndarray: The diagonal of the assembled tangent matrix.
  """
  n_fields = static_settings['number of fields'][set]
  user_residual_fun = static_settings['model'][set]
  connectivity_list = jnp.asarray(static_settings['connectivity'][set])
  x_nodes = settings['node coordinates']
  local_dofs = dofs[connectivity_list]
  local_node_coor = x_nodes[connectivity_list]
  elem_numbers = jnp.arange(0, connectivity_list.shape[0], 1)

  # For each element call user element routine (as the one in pde-module with mode='residual') and get contributions
  tangent_contributions = jax.vmap(jax.jacfwd(user_residual_fun), (0, 0, 0, None, None, None), (0))(local_dofs, local_node_coor, elem_numbers, settings, static_settings, set)
  diagonal_contributions = jnp.diagonal(tangent_contributions, axis1=1, axis2=2)

  # Assemble diagonal entries like a residual
  flat_neighbors = connectivity_list.flatten()
  tmp = jnp.outer(jnp.ones(flat_neighbors.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_))
  indices = tmp + jnp.outer(flat_neighbors, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)
  diag = jnp.zeros(dofs.shape).flatten()
  diag = diag.at[indices.flatten()].add(diagonal_contributions.flatten())
  return diag

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def user_residual_assemble_tangent(dofs, settings, static_settings, set):
  """
  Assembly of the full (sparse) tangent matrix for custom user residual of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jax.experimental.sparse.BCOO: The assembled tangent matrix.
  """
  n_fields = static_settings['number of fields'][set]
  user_residual_fun = static_settings['model'][set]
  connectivity_list = jnp.asarray(static_settings['connectivity'][set])
  x_nodes = settings['node coordinates']
  local_dofs = dofs[connectivity_list]
  local_node_coor = x_nodes[connectivity_list]
  dofs_per_elem = local_dofs[0].flatten().shape[0]
  elem_numbers = jnp.arange(0, connectivity_list.shape[0], 1)

  # For each element call user element routine (as the one in pde-module with mode='residual') and get contributions
  tangent_contributions = jax.vmap(jax.jacfwd(user_residual_fun), (0, 0, 0, None, None, None), (0)
                                   )(local_dofs, local_node_coor, elem_numbers, settings, static_settings, set)

  # Assembling (without summing duplicates)
  def get_indices(list):
    def one_elem_idx(neighb):
      def row_indices(a, row):
        return jnp.asarray([[a,row[i]] for i in range(row.shape[0])])
      colrow = jax.vmap(row_indices, (0, None), 0)
      tmp = jnp.outer(jnp.ones(neighb.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_)).astype(jnp.int_)
      indices = tmp + jnp.outer(neighb, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)
      flat_indices = indices.flatten()
      return colrow(flat_indices,flat_indices)
    all_elem_idx = jax.vmap(one_elem_idx, (0), 0)
    return jnp.reshape(all_elem_idx(list), shape=(list.shape[0]*(dofs_per_elem)**2,2))
  
  indices = get_indices(connectivity_list)
  data = tangent_contributions.flatten()
  n_dofs = n_fields * dofs.shape[0]

  tangent = sparse.BCOO((data, indices), shape=(n_dofs,n_dofs))
  return tangent


### Assembling for user elements
@jit_with_docstring(static_argnames=['static_settings', 'set'])
def user_element_assemble_residual(dofs, settings, static_settings, set):
  """
  Assembly of residual for custom user element of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jnp.ndarray: The assembled residual.
  """

  n_fields = static_settings['number of fields'][set]
  user_elem_fun = static_settings['model'][set]
  connectivity_list = jnp.asarray(static_settings['connectivity'][set])
  x_nodes = settings['node coordinates']
  local_dofs = dofs[connectivity_list]
  local_node_coor = x_nodes[connectivity_list]
  elem_numbers = jnp.arange(0, connectivity_list.shape[0], 1)


  # For each element call user element routine (as the one in pde-module with mode='residual') and get contributions
  residual_contributions = jax.vmap(user_elem_fun, (0, 0, 0, None, None, None, None), (0))(local_dofs, local_node_coor, elem_numbers, settings, static_settings, 'residual', set)

  # Assembling
  flat_neighbors = connectivity_list.flatten()
  tmp = jnp.outer(jnp.ones(flat_neighbors.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_))
  indices = tmp + jnp.outer(flat_neighbors, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)

  res = jnp.zeros(dofs.shape).flatten()
  res = res.at[indices.flatten()].add(residual_contributions.flatten())
  res = res.reshape(dofs.shape)
  res = jnp.reshape(res, dofs.shape)

  return res

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def user_element_assemble_tangent_diagonal(dofs, settings, static_settings, set):
  """
  Assembly of the diagonal of the tangent matrix for custom user element of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jnp.ndarray: The diagonal of the assembled tangent matrix.
  """

  n_fields = static_settings['number of fields'][set]
  user_elem_fun = static_settings['model'][set]
  connectivity_list = jnp.asarray(static_settings['connectivity'][set])
  x_nodes = settings['node coordinates']
  local_dofs = dofs[connectivity_list]
  local_node_coor = x_nodes[connectivity_list]
  elem_numbers = jnp.arange(0, connectivity_list.shape[0], 1)

  # For each element call user element routine (as the one in pde-module with mode='residual') and get contributions
  tangent_contributions = jax.vmap(user_elem_fun, (0, 0, 0, None, None, None, None), (0))(local_dofs, local_node_coor, elem_numbers, settings, static_settings, 'tangent', set)
  diagonal_contributions = jnp.diagonal(tangent_contributions, axis1=1, axis2=2)

  # Assemble diagonal entries like a residual
  flat_neighbors = connectivity_list.flatten()
  tmp = jnp.outer(jnp.ones(flat_neighbors.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_))
  indices = tmp + jnp.outer(flat_neighbors, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)
  diag = jnp.zeros(dofs.shape).flatten()
  diag = diag.at[indices.flatten()].add(diagonal_contributions.flatten())
  return diag

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def user_element_assemble_tangent(dofs, settings, static_settings, set):
  """
  Assembly of the full (sparse) tangent matrix for custom user element of specified set.
  
  Parameters:
    dofs (jnp.ndarray): Degrees of freedom.
    settings (dict): Settings for the integration.
    static_settings (flax.core.FrozenDict): Static settings including assembling modes.
    set (int): The set index.
  
  Returns:
    jax.experimental.sparse.BCOO: The assembled tangent matrix.
  """
  
  n_fields = static_settings['number of fields'][set]
  user_elem_fun = static_settings['model'][set]
  connectivity_list = jnp.asarray(static_settings['connectivity'][set])
  x_nodes = settings['node coordinates']
  local_dofs = dofs[connectivity_list]
  local_node_coor = x_nodes[connectivity_list]
  dofs_per_elem = local_dofs[0].flatten().shape[0]
  elem_numbers = jnp.arange(0, connectivity_list.shape[0], 1)

  # For each element call user element routine (as the one in pde-module with mode='residual') and get contributions
  tangent_contributions = jax.vmap(user_elem_fun, (0, 0, 0, None, None, None, None), (0)
                                   )(local_dofs, local_node_coor, elem_numbers, settings, static_settings, 'tangent', set)

  # Assembling (without summing duplicates)
  def get_indices(list):
    def one_elem_idx(neighb):
      def row_indices(a, row):
        return jnp.asarray([[a,row[i]] for i in range(row.shape[0])])
      colrow = jax.vmap(row_indices, (0, None), 0)
      tmp = jnp.outer(jnp.ones(neighb.shape, dtype=jnp.int_), jnp.arange(n_fields, dtype=jnp.int_)).astype(jnp.int_)
      indices = tmp + jnp.outer(neighb, n_fields * jnp.ones(n_fields, dtype=jnp.int_)).astype(jnp.int_)
      flat_indices = indices.flatten()
      return colrow(flat_indices,flat_indices)
    all_elem_idx = jax.vmap(one_elem_idx, (0), 0)
    return jnp.reshape(all_elem_idx(list), shape=(list.shape[0]*(dofs_per_elem)**2,2))
  
  indices = get_indices(connectivity_list)
  data = tangent_contributions.flatten()
  n_dofs = n_fields * dofs.shape[0]

  tangent = sparse.BCOO((data, indices), shape=(n_dofs,n_dofs))
  return tangent



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import time
import sys
import os

import jax
from jax import config
config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')

import jax.numpy as jnp
import flax

from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models


### First order system of stsationary heat conduction, mls solution structure, dirichlet/neumann/robin conditions


# Positive smooth distance functions that select the specific boundary segment
# Time dimension is always the last dimension (here the second)
psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, [0.,0.], [1.,0.])
psdf_right = lambda x: geometry.psdf_trimmed_line(x, [1.,0.], [1.,1.])
psdf_left = lambda x: geometry.psdf_trimmed_line(x, [0.,1.], [0.,0.])
psdfs = [psdf_bottom, psdf_right, psdf_left]

# Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
psdf_rect = lambda x: geometry.psdf_parallelogram(x, [0.,0.], [0.,1.], [1.,0.])
psdf_domain = psdf_rect

# Initial condition
def bc_bottom(x):
  return 0.5 * (1 - jax.lax.cos(2 * jnp.pi * x[0]))

# Dirichlet boundary conditions on left an right side
def bc_left(x):
  return 0.
def bc_right(x):
  return 0.
dirichlet_conditions = [bc_bottom, bc_right, bc_left]

# Field selector for boundary conditions
def dirichlet(x):
  return jnp.asarray([1., 0.])  # 1 * theta + 0 * dtheta / dx = bc_bottom
def neumann(x):
  return jnp.asarray([0., 1.]) # 0 * theta + 1 * dtheta / dx = bc_right
def robin(x):
  return jnp.asarray([0.980581, -0.196116]) # 0.980581 * theta - 0.196116 * dtheta / dx = bc_left
a_i = [dirichlet, neumann, robin]

# Transfinite interpolation of boundary segments
b_coeff = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, a_i)])
b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, dirichlet_conditions)])
psdf = lambda x: jnp.asarray([geometry.psdf_unification(x, psdfs)])

### Initialization and generation of nodes and integration points
nod_spacing = 0.01
(x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_domain, [0.,0.], [1.,1.], nod_spacing)

# Collocation
x_int = x_nodes
n_int = n_nodes
w_int = (region_size / n_int) * jnp.ones(n_int)
print("Estimated region size: ", w_int.sum())

# Support radius for moving least square ansatz
support_radius = 2.51 * nod_spacing

# Definition of pde
diffusivity = lambda x: 0.1
pde = models.heat_equation_fos(diffusivity_fun=diffusivity)

### Neighborhood search
(num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)


### Settings
n_fields = 2
dofs0 = jnp.zeros((n_nodes, n_fields))
static_settings = flax.core.FrozenDict({
  'solution space': ('mls',),
  'shape function mode': 'compiled',
  'number of fields': (n_fields,),
  'assembling mode': ('sparse',),
  'maximal number of neighbors': (max_neighbors,),
  'order of basis functions': (2,),
  'weight function type': ('gaussian',),
  'variational scheme': ('least square pde loss',),
  'model': (pde,),
  'solution structure': ('first order set',),
  'boundary coefficients': (b_coeff,),
  'boundary conditions': (b_cond,),
  'psdf': (psdf,),
  'solver type': 'linear',
  'solver backend': 'pardiso',
  'solver': 'lu',
  'verbose': 1,
  'connectivity': (utility.jnp_to_tuple(neighbor_list),),
})
settings = {
  'beta': (3,),
  'node coordinates': x_nodes,
  'integration coordinates': (x_int,),
  'integration weights': (w_int,),
  'support radius': (support_radius,),
}

# Precompute shape functions and pass them to e.g. static_settings...
settings = solution_structures.precompile(dofs0, settings, static_settings)

# Call solver
dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8)

# Preperation of postprocessing data
local_dofs = dofs[neighbor_list]
def post_fun(x, itt, local_dofs, settings, static_settings):
  return solution_structures.solution_structure(x, itt, local_dofs, settings, static_settings, 0)

post_fun_vj = jax.jit(jax.vmap(post_fun, (0,0,0,None,None)), static_argnames=['static_settings'])
data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings)

# Plotting with pyvista
plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True)



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

import time
import sys

import jax
from jax import lax, config
import jax.numpy as jnp
import flax

from autopdex import seeder, geometry, solver, solution_structures, plotter, utility, models, spaces, assembler

config.update("jax_enable_x64", True)
config.update("jax_compilation_cache_dir", './cache')


### First order least square Laplace equation with coupled Neumann boundary conditions on circular cutout



# Positive smooth distance functions that select the specific boundary segments
psdf_bottom = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,0.]), jnp.stack([1.,0.]))
psdf_right = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([1.,0.]), jnp.stack([1.,1.]))
psdf_top = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([1.,1.]), jnp.stack([0.,1.]))
psdf_left = lambda x: geometry.psdf_trimmed_line(x, jnp.stack([0.,1.]), jnp.stack([0.,0.]))
psdf_cutout = lambda x: geometry.psdf_nd_sphere_cutout(x, jnp.stack([0.5,0.5]), 0.25)
psdf_rect = lambda x: geometry.psdf_parallelogram(x, jnp.stack([0.,0.]), jnp.stack([0.,1.]), jnp.stack([1.,0.]))
psdfs = [psdf_bottom, psdf_right, psdf_top, psdf_left, psdf_cutout]

# Positive smooth distance funcion of full domain (used for seeding nodes and integration points)
psdf_domain = lambda x: geometry.r_equivalence(psdf_rect(x), psdf_cutout(x))

# Boundary condition functions
def homogeneous(x): # Bottom, right and top boundaries
  return 0
def bc_0(x): # Left boundary
  return jax.lax.sin(jnp.pi * x[0])
def neumann_at_circ(x): # Normal derivative at circular cutout for a specific analytic solution, may be used for convergence study
  return (2. * jnp.pi / jnp.sinh(jnp.pi)) * ((-1.+2.*x[1])*jnp.cosh(jnp.pi-jnp.pi*x[1]) + (1.-2.*x[0])*jnp.cos(jnp.pi*x[0])*jnp.sinh(jnp.pi-jnp.pi*x[1]))

# Field selectors for boundary conditions
def dirichlet(x):
  return jnp.asarray([1., 0., 0.])
def coupled_neumann(x):
  centroid = jnp.asarray([0.5,0.5])
  dxc = x - centroid
  normal = - dxc / jnp.sqrt(jnp.dot(dxc, dxc))
  return jnp.concatenate([jnp.asarray([0.]), normal])

# Assignment of boundary conditions
dirichlet_conditions = [bc_0, homogeneous, homogeneous, homogeneous, neumann_at_circ]
a_i = [dirichlet, dirichlet, dirichlet, dirichlet, coupled_neumann]

### Transfinite interpolation of boundary segments
b_coeff = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, a_i)])
b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, dirichlet_conditions)])
psdf = lambda x: jnp.asarray([psdf_domain(x)])



### Initialization and generation of nodes and integration points
# Regular node placement
nod_spacing = 0.025
(x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(psdf_domain, [0.,0.], [1.,1.], nod_spacing)

# Integration points
# # Regular spacing
# int_spacing = nod_spacing
# (x_int, w_int, n_int) = seeder.gauss_points_in_psdf(psdf_domain, [0.,0.], [1.,1.], int_spacing, order=2)
# Collocation
x_int = x_nodes
n_int = n_nodes
w_int = (region_size / n_int) * jnp.ones(n_int)
print("Estimated region size: ", w_int.sum())



# Support radius for moving least square ansatz
support_radius = 2.51 * nod_spacing

# Definition of pde
pde = models.poisson_fos(support_radius)

# Neighborhood search
(num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)



### Settings
n_fields = 3
dofs0 = jnp.zeros((n_nodes, n_fields))
static_settings = flax.core.FrozenDict({
  'solution space': ('mls',),
  'shape function mode': 'compiled',
  'number of fields': (n_fields,),
  'assembling mode': ('sparse',),
  'maximal number of neighbors': (max_neighbors,),
  'order of basis functions': (1,),
  'weight function type': ('gaussian',),
  'variational scheme': ('least square pde loss',),
  'model': (pde,),
  'solution structure': ('first order set',),
  'boundary coefficients': (b_coeff,),
  'boundary conditions': (b_cond,),
  'psdf': (psdf,),
  'solver type': 'linear',
  'solver backend': 'pardiso',
  'solver': 'lu',
  'verbose': 1,
  'connectivity': (utility.jnp_to_tuple(neighbor_list),),
})
settings = {
  'beta': (3,),
  'node coordinates': x_nodes,
  'integration coordinates': (x_int,),
  'integration weights': (w_int,),
  'support radius': (support_radius,),
}

### Optional: Precompute shape functions and pass them to e.g. static_settings...
settings = solution_structures.precompile(dofs0, settings, static_settings)


### Call solver
dofs, _ = solver.solver(dofs0, settings, static_settings, tol=1e-8)


### Preperation of postprocessing data
local_dofs = dofs[neighbor_list]
def post_fun(x, itt, local_dofs, settings, static_settings):
  return solution_structures.solution_structure(x, itt, local_dofs, settings, static_settings, 0)

post_fun_vj = jax.jit(jax.vmap(post_fun, (0,0,0,None,None)), static_argnames=['static_settings'])
data = post_fun_vj(x_int, jnp.arange(n_int), local_dofs, settings, static_settings)

# Plotting with pyvista
plotter.pv_plot(x_int, data[:,0], export_vtk=False, show=True)



# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

def test_example_circle_in_channel():
  from jax import config
  import jax.numpy as jnp
  import flax

  from autopdex import seeder, geometry, solution_structures, utility, models, assembler

  config.update("jax_enable_x64", True)

  # Domain selection
  channel = lambda x: geometry.sdf_parallelogram_2d(x, [-6.732,-1.], [6.732,-1.], [-6.732,1.])
  outer_circle = lambda x: geometry.sdf_nd_sphere(x, [0.,0.], 2.)
  inner_circle_cut = lambda x: -geometry.sdf_nd_sphere(x, [0.,0.], 1.)
  outer_and_channel = lambda x: geometry.r_disjunction(channel(x), outer_circle(x))
  sdf_domain = lambda x: geometry.r_conjunction(outer_and_channel(x), inner_circle_cut(x))

  # Smooth distance functions that select the specific boundary segments
  psdf_left = lambda x: geometry.psdf_trimmed_line(x, [-6.732,-1.], [-6.732,1.])
  psdf_right = lambda x: geometry.psdf_trimmed_line(x, [6.732,-1.], [6.732,1.])
  channel_2 = lambda x: geometry.sdf_nd_planes(x, [-6.732,-1.], [-6.732,1.])
  outer_and_channel_2 = lambda x: geometry.r_disjunction(channel_2(x), outer_circle(x))
  sdf_remaining = lambda x: geometry.r_conjunction(outer_and_channel_2(x), inner_circle_cut(x))
  psdfs = [psdf_left, psdf_right, sdf_remaining]

  # Smooth distance funcion of full domain
  sdf = lambda x: jnp.asarray([sdf_domain(x), sdf_domain(x)])

  # Boundary condition functions
  homogeneous = lambda x: 0.

  # Notice the different flow field for changing inlet velocity
  v_left = lambda x: 0.5 * (1 - x[1]**2)
  # v_left = lambda x: 0.01 * (1 - x[1]**2)

  # Field selectors for boundary conditions
  selection_dirichlet_1 = lambda x: jnp.asarray([0.,1.,0.])
  selection_dirichlet_2 = lambda x: jnp.asarray([0.,0.,1.])
  selection_pressure = lambda x: jnp.asarray([1.,0.,0.])

  # Assignment of boundary conditions (same ordering as in psdfs)
  bc_1 = [v_left, homogeneous, homogeneous] # first boundary condition at left, right and remaining boundaries
  bc_2 = [homogeneous, homogeneous, homogeneous] # second bc

  # Selectors that define for which field the boundary conditions shall be imposed
  ai_1 = [selection_dirichlet_1, selection_pressure, selection_dirichlet_1]
  ai_2 = [selection_dirichlet_2, selection_dirichlet_2, selection_dirichlet_2]

  ### Transfinite interpolation of boundary segments
  b_coeff = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, ai_1),
                                  geometry.transfinite_interpolation(x, psdfs, ai_2)])
  b_cond = lambda x: jnp.asarray([geometry.transfinite_interpolation(x, psdfs, bc_1),
                                  geometry.transfinite_interpolation(x, psdfs, bc_2)])


  ### Initialization and generation of nodes and integration points
  # Regular node placement
  nod_spacing = 1.
  # nod_spacing = 0.05
  (x_nodes, n_nodes, region_size) = seeder.regular_in_psdf(sdf_domain, [-6.732,-2.], [6.732,2.], nod_spacing, atol=-1e-8)

  # Gauss integration on background grid
  int_spacing = nod_spacing
  (x_int, w_int, n_int) = seeder.gauss_points_in_psdf(sdf_domain, [-6.732,-2.], [6.732,2.], int_spacing, order=2)

  # Support radius for moving least square ansatz
  support_radius = 3.01 * nod_spacing

  # Definition of pde
  pde = models.navier_stokes_incompressible_steady(dynamic_viscosity=0.01, incompressible_weighting=10.)

  ### Neighborhood search
  (num_neighbors, max_neighbors, min_neighbors, neighbor_list) = utility.search_neighborhood(x_nodes, x_int, support_radius)

  ### Setting
  n_fields = 3
  dofs0 = jnp.zeros((n_nodes, n_fields))
  static_settings = flax.core.FrozenDict({
    'solution space': ('mls',),
    'shape function mode': 'compiled',
    'number of fields': (n_fields,),
    'assembling mode': ('sparse',),
    'maximal number of neighbors': (max_neighbors,),
    'order of basis functions': (2,),
    'weight function type': ('gaussian',),

    # The least square variational approach minimizes a functional with respect to the 
    # degrees of freedom. As a consequence, also nonlinear minimizers can be utilized.
    'variational scheme': ('least square pde loss',),
    # 'solver type': 'minimize',
    # 'solver': 'lbfgs',
    'solver type': 'newton',
    'solver backend': 'scipy',
    'solver': 'lapack',

    'model': (pde,),
    'solution structure': ('second order set',),
    'boundary coefficients': (b_coeff,),
    'boundary conditions': (b_cond,),
    'psdf': (sdf,),
    'verbose': 1,
    'connectivity': (utility.jnp_to_tuple(neighbor_list),),
  })
  settings = {
    'beta': (3.,),
    'node coordinates': x_nodes,
    'integration coordinates': (x_int,),
    'integration weights': (w_int,),
    'support radius': (support_radius,),
  }


  # Precompute shape functions
  settings = solution_structures.precompile(dofs0, settings, static_settings)

  tangent = assembler.assemble_tangent(dofs0, settings, static_settings)
  check = tangent.data.flatten().sum()
#   print(check)
  assert jnp.isclose(check, 1427.552335434657)



# variational_schemes.py
# Copyright (C) 2024 Tobias Bode
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.

"""
This module contains functions for computing residuals and functionals at integration points using different variational schemes.

Usually, the variational_schemes module does not have to be called by the user, but it is invoked by the assembly module in the 
'sparse' and 'dense' modes and evaluates integration point contributions of the functional to be integrated or the residual to be assembled. 
A distinction is made between the least squares variational method for solving PDEs or 
approximating functions and the Galerkin method in its weak and strong (without integration by parts) forms.
"""

import jax
import jax.numpy as jnp

from autopdex import spaces
from autopdex import solution_structures as ss
from autopdex.utility import jit_with_docstring



@jit_with_docstring(static_argnames=['static_settings', 'set'])
def functional_at_int_point(x_i, w_i, int_point_number, local_dofs, settings, static_settings, set):
  """
  Computes the functional at an integration point for 'least square pde loss' or 'least square function approximation'. 

  Args:
    x_i (jnp.ndarray): Coordinates of the integration point.
    w_i (float): Integration weight.
    int_point_number (int): Integration point number.
    local_dofs (jnp.ndarray): Local degrees of freedom.
    settings (dict): Settings for the computation.
    static_settings (dict): Static settings for the computation.
    set (str): Identifier for the current set.

  Returns:
    float: Computed functional value.
  """
  algorithm = static_settings['variational scheme'][set]
  match algorithm:
    case 'least square function approximation':
      return least_square_function_approximation(x_i, w_i, int_point_number, local_dofs, settings, static_settings, set)
    case 'least square pde loss':
      return least_square_pde_loss(x_i, w_i, int_point_number, local_dofs, settings, static_settings, set)
    case _:
      assert False, 'variational scheme is not defined or not available with this settings.'

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def direct_residual_at_int_point(x_i, w_i, int_point_number, local_dofs, settings, static_settings, set):
  """
  Computes the direct residual at an integration point for 'strong form galerkin'.

  Args:
    x_i (jnp.ndarray): Coordinates of the integration point.
    w_i (float): Integration weight.
    int_point_number (int): Integration point number.
    local_dofs (jnp.ndarray): Local degrees of freedom.
    settings (dict): Settings for the computation.
    static_settings (dict): Static settings for the computation.
    set (str): Identifier for the current set.

  Returns:
    float: Computed residual value.
  """
  algorithm = static_settings['variational scheme'][set]
  match algorithm:
    case 'strong form galerkin':
      return strong_form_galerkin(x_i, w_i, int_point_number, local_dofs, settings, static_settings, set)
    case _:
      assert False, 'variational scheme is not defined or not available with this settings.'

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def residual_from_deriv_at_int_point(x_i, w_i, int_point_number, local_dofs, local_virt_dofs, settings, static_settings, set):
  """
  Computes the residual from derivatives at an integration point for 'weak form galerkin'.

  Args:
    x_i (jnp.ndarray): Coordinates of the integration point.
    w_i (float): Integration weight.
    int_point_number (int): Integration point number.
    local_dofs (jnp.ndarray): Local degrees of freedom.
    local_virt_dofs (jnp.ndarray): Local virtual degrees of freedom.
    settings (dict): Settings for the computation.
    static_settings (dict): Static settings for the computation.
    set (str): Identifier for the current set.

  Returns:
    float: Computed residual value.
  """
  algorithm = static_settings['variational scheme'][set]
  match algorithm:
    case 'weak form galerkin':
      return weak_form_galerkin(x_i, w_i, int_point_number, local_dofs, local_virt_dofs, settings, static_settings, set)
    case _:
      assert False, 'variational scheme is not defined or not available with this settings.'

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def least_square_pde_loss(x_i, w_i, int_point_number, local_dofs, settings, static_settings, set):
  """
  Computes the least square PDE loss at an integration point.

  Takes the set of pdes from static_settings['model']

  Args:
    x_i (jnp.ndarray): Coordinates of the integration point.
    w_i (float): Integration weight.
    int_point_number (int): Integration point number.
    local_dofs (jnp.ndarray): Local degrees of freedom.
    settings (dict): Settings for the computation.
    static_settings (dict): Static settings for the computation.
    set (str): Identifier for the current set.

  Returns:
    float: Computed least square PDE loss.
  """
  def approximate_solution(x):
    ansatz = ss.solution_structure(x, int_point_number, local_dofs, settings, static_settings, set)

    if ansatz.shape[0] == 1:
      return ansatz[0]
    else:
      return ansatz

  if static_settings['shape function mode'] == 'compiled':
    x_i = 0*x_i

  pde = static_settings['model'][set](x_i, approximate_solution, settings, static_settings, int_point_number, set)

  return w_i * jnp.dot(pde, pde) / 2

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def weak_form_galerkin(x_i, w_i, int_point_number, local_dofs, local_virt_dofs, settings, static_settings, set):
  """
  Computes the weak form Galerkin residual at an integration point.

  Takes the set of pdes from static_settings['model']

  Args:
    x_i (jnp.ndarray): Coordinates of the integration point.
    w_i (float): Integration weight.
    int_point_number (int): Integration point number.
    local_dofs (jnp.ndarray): Local degrees of freedom.
    local_virt_dofs (jnp.ndarray): Local virtual degrees of freedom.
    settings (dict): Settings for the computation.
    static_settings (dict): Static settings for the computation.
    set (str): Identifier for the current set.

  Returns:
    float: Computed weak form Galerkin residual.
  """
  if static_settings['shape function mode'] == 'compiled':
    x_i = 0*x_i

  def weak_form(virtual_dofs):
    def test_function(x):
      if local_dofs.shape[1] == 1:
        return ss.solution_structure(x, int_point_number, virtual_dofs, settings, static_settings, set)[0]
      else:
        return ss.solution_structure(x, int_point_number, virtual_dofs, settings, static_settings, set)
    def trial_function(x):
      if local_dofs.shape[1] == 1:
        return ss.solution_structure(x, int_point_number, local_dofs, settings, static_settings, set)[0]
      else:
        return ss.solution_structure(x, int_point_number, local_dofs, settings, static_settings, set)

    # Evaluate weak form and take derivative with respect to local_virtual_dofs
    weak_pde = static_settings['model'][set](x_i, trial_function, test_function, settings, static_settings, int_point_number, set)
    return weak_pde

  # Test performance for jacrev and jacfwd
  residual = w_i * jax.jacrev(weak_form)(local_virt_dofs)
  return residual

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def strong_form_galerkin(x_i, w_i, int_point_number, local_dofs, settings, static_settings, set):
  """
  Computes the strong form Galerkin residual at an integration point.

  Takes the set of pdes from static_settings['model']

  Args:
      x_i (jnp.ndarray): Coordinates of the integration point.
    w_i (float): Integration weight.
    int_point_number (int): Integration point number.
    local_dofs (jnp.ndarray): Local degrees of freedom.
    settings (dict): Settings for the computation.
    static_settings (dict): Static settings for the computation.
    set (str): Identifier for the current set.

  Returns:
    float: Computed strong form Galerkin residual.
  """
  def approximate_solution(x):
    if local_dofs.shape[1] == 1:
      return ss.solution_structure(x, int_point_number, local_dofs, settings, static_settings, set)[0]
    else:
      return ss.solution_structure(x, int_point_number, local_dofs, settings, static_settings, set)

  shp_fun_mode = static_settings['shape function mode']
  if shp_fun_mode == 'compiled':
    x_i = 0*x_i
    test_functions = settings['compiled shape functions'][set][0][int_point_number]
  elif shp_fun_mode == 'direct':
    test_functions = jax.jacrev(spaces.solution_space, argnums=2)(x_i, int_point_number, local_dofs, settings, static_settings, set)[0,:,0]
  else:
    assert False, 'Shape function mode neither specified as \'direct\' nor \'compiled\'!'

  pde = static_settings['model'][set](x_i, approximate_solution, settings, static_settings, int_point_number, set)

  if local_dofs.shape[1] == 1:
    return w_i * test_functions * pde
  else:
    return w_i * jnp.outer(test_functions, pde)

@jit_with_docstring(static_argnames=['static_settings', 'set'])
def least_square_function_approximation(x_i, w_i, int_point_number, local_dofs, settings, static_settings, set):  
  """
  Computes the least square function approximation at an integration point.

  Takes the set of functions to fit from static_settings['analytic solution'].
  
  Args:
    x_i (jnp.ndarray): Coordinates of the integration point.
    w_i (float): Integration weight.
    int_point_number (int): Integration point number.
    local_dofs (jnp.ndarray): Local degrees of freedom.
    settings (dict): Settings for the computation.
    static_settings (dict): Static settings for the computation.
    set (str): Identifier for the current set.

  Returns:
    float: Computed least square function approximation.
  """
  def approximate_solution(x):
    return ss.solution_structure(x, int_point_number, local_dofs, settings, static_settings, set)

  analytic_solution = static_settings['analytic solution'][set]

  # L2 error
  error = analytic_solution(x_i) - approximate_solution(x_i)
  return w_i * jnp.dot(error, error) / 2
